@@ -1,6 +1,6 @@
 # KInterbasDB Python Package - Python Wrapper for Core
 #
-# Version 3.2
+# Version 3.3
 #
 # The following contributors hold Copyright (C) over their respective
 # portions of code (see license.txt for details):
@@ -23,87 +23,97 @@
 # relied upon not to change.  Module members whose names begin with an
 # underscore cannot be expected to have stable interfaces.
 
-__version__ = (3, 2, 0, 'final', 0)
-__timestamp__ = '2006.08.11.15.17.16.UTC'
+__version__ = (3, 3, 0, 'pre-alpha', 0)
+__timestamp__ = '2009.01.13.11.52.28.UTC'
 
 import os, struct, sys
 
+_FS_ENCODING = sys.getfilesystemencoding()
+
 if sys.platform.lower().startswith('win'):
     import os.path
 
     # Better out-of-box support for embedded DB engine on Windows:  if the
     # client library is detected in the same directory as kinterbasdb, or in
     # the 'embedded' subdirectory of kinterbasdb's directory, or in the same
-    # directory as the Python executable, give that precedence over the
-    # location listed in the registry.
+    # directory as the Python executable, give that library instance precedence
+    # over the location listed in the registry.
     #
-    # Client programmers who happen to read this should note that the embedded
-    # engine malfunctions in various ways unless it's located in the same
-    # directory as the actual executable (python.exe).
-    _clientLibDir = None
-    _kinterbasdbDir = os.path.dirname(os.path.abspath(__file__))
-    for _clientLibName in (
-        'firebird.dll', # Vulcan
-        'fbclient.dll', # FB 1.5, 2.x
-      ):
-        for _location in (
-            os.path.join(os.getcwd(), _clientLibName),
-            os.path.join(_kinterbasdbDir, _clientLibName),
-            os.path.join(os.path.join(_kinterbasdbDir, 'embedded'),
-                _clientLibName
-              ),
-            os.path.join(os.path.dirname(sys.executable), _clientLibName),
+    # 2007.03.19: Overhauled to support non-ASCII paths properly, in response
+    # to:
+    #   http://sourceforge.net/forum/forum.php?thread_id=1695175&forum_id=30917
+    def _findAndLoadFBDLLDir():
+        clientLibDir = None
+        curWorkingDir = os.getcwdu()
+        kinterbasdbDir = os.path.dirname(os.path.abspath(__file__)).decode(
+            _FS_ENCODING
+          )
+        pythonDir = os.path.dirname(sys.executable).decode(_FS_ENCODING)
+        for clientLibName in (
+            u'firebird.dll', # Vulcan
+            u'fbclient.dll', # FB 1.5, 2.x
           ):
-            if os.path.isfile(_location):
-                _clientLibDir = os.path.dirname(_location)
+            for location in (
+                os.path.join(curWorkingDir, clientLibName),
+                os.path.join(kinterbasdbDir, clientLibName),
+                os.path.join(os.path.join(kinterbasdbDir, u'embedded'),
+                    clientLibName
+                  ),
+                os.path.join(pythonDir, clientLibName),
+              ):
+                if os.path.isfile(location):
+                    clientLibDir = os.path.dirname(location)
+                    break
+
+            if clientLibDir:
                 break
 
-        if _clientLibDir:
-            break
+        if clientLibDir:
+            origOSPath = os.environ['PATH'].decode(_FS_ENCODING)
+            os.environ['PATH'] = (
+                origOSPath +  os.pathsep + clientLibDir
+              ).encode(_FS_ENCODING)
 
-    if _clientLibDir:
-        os.environ['PATH'] = _clientLibDir + os.pathsep + os.environ['PATH']
-        # At least with FB 1.5.2, the FIREBIRD environment variable must also
-        # be set in order for all features to work properly.
-        os.environ['FIREBIRD'] = _clientLibDir
-    else:
-        # FB 1.5 RC7 and later, when installed via the packaged installer or
-        # the "instreg.exe" command-line tool, record their installation dir in
-        # the registry.  If no client library was detected earlier, we'll add
-        # the "bin" subdirectory of the directory from the registry to the
-        # *end* of the PATH, so it'll be used as a last resort.
-        import _winreg
+            # At least with FB 1.5.2, the FIREBIRD environment variable must
+            # also be set in order for all features to work properly.
+            os.environ['FIREBIRD'] = clientLibDir.encode(_FS_ENCODING)
+        else:
+            # FB 1.5 RC7 and later, when installed via the packaged installer
+            # or the "instreg.exe" command-line tool, record their installation
+            # dir in the registry.  If no client library was detected earlier,
+            # we'll add the "bin" subdirectory of the directory from the
+            # registry to the *end* of the PATH, so it'll be used as a last
+            # resort.
+            import _winreg
 
-        _reg = _winreg.ConnectRegistry(None, _winreg.HKEY_LOCAL_MACHINE)
-        try:
+            reg = _winreg.ConnectRegistry(None, _winreg.HKEY_LOCAL_MACHINE)
             try:
-                _dbInstPathsKey = _winreg.OpenKey(
-                    _reg, r'SOFTWARE\Firebird Project\Firebird Server'
-                           '\Instances'
-                  )
                 try:
-                    _instPath = _winreg.QueryValueEx(
-                        _dbInstPathsKey, 'DefaultInstance'
-                      )[0]
-                finally:
-                    _dbInstPathsKey.Close()
-                    del _dbInstPathsKey
-            except WindowsError:
-                # Versions of IB/FB prior to FB 1.5 RC7 don't have this _reg
-                # entry, but they install the client library into a system
-                # library directory, so there's no problem.
-                pass
-            else:
-                os.environ['PATH'] += os.pathsep + os.path.join(
-                    _instPath, 'bin'
-                  )
-                del _instPath
-        finally:
-            _reg.Close()
+                    dbInstPathsKey = _winreg.OpenKey(reg,
+                        u'SOFTWARE\\Firebird Project\\Firebird Server'
+                        u'\\Instances'
+                      )
+                    try:
+                        instPath = _winreg.QueryValueEx(
+                            dbInstPathsKey, u'DefaultInstance'
+                          )[0]
+                    finally:
+                        dbInstPathsKey.Close()
+                except WindowsError:
+                    # Versions of IB/FB prior to FB 1.5 RC7 don't have this reg
+                    # entry, but they install the client library into a system
+                    # library directory, so there's no problem.
+                    pass
+                else:
+                    origOSPath = os.environ['PATH'].decode(_FS_ENCODING)
+                    fbBinPath = os.pathsep + os.path.join(instPath, u'bin')
+                    os.environ['PATH'] = (origOSPath + fbBinPath).encode(
+                        _FS_ENCODING
+                      )
+            finally:
+                reg.Close()
 
-        del _winreg, _reg
-
-    del _clientLibDir, _kinterbasdbDir, _clientLibName, _location
+    _findAndLoadFBDLLDir()
 
 # The underlying C module:
 import _kinterbasdb as _k
@@ -135,7 +145,8 @@ IntegrityError        = k_exceptions.IntegrityError     = _k.IntegrityError
 InternalError         = k_exceptions.InternalError      = _k.InternalError
 
 ProgrammingError      = k_exceptions.ProgrammingError   = _k.ProgrammingError
-TransactionConflict          = k_exceptions.TransactionConflict       = _k.TransactionConflict
+TransactionConflict \
+  = k_exceptions.TransactionConflict = _k.TransactionConflict
 
 NotSupportedError     = k_exceptions.NotSupportedError  = _k.NotSupportedError
 
@@ -206,7 +217,7 @@ default_tpb = (
     + isc_tpb_write                                 # Access mode
     + isc_tpb_read_committed + isc_tpb_rec_version  # Isolation level
     + isc_tpb_wait                                  # Lock resolution strategy
-    + isc_tpb_shared                                # Table reservation
+#    + isc_tpb_shared                                # Table reservation
                                                     # access method
   )
 
@@ -250,9 +261,11 @@ _NORMAL_TYPE_TRANS_OUT = None
 
 
 initialized = False
-def init(type_conv=1, concurrency_level=_k.DEFAULT_CONCURRENCY_LEVEL):
+_guessTextualBlobEncodingWhenUsingFB20AndEarlier = False
+def init(type_conv=200, concurrency_level=_k.DEFAULT_CONCURRENCY_LEVEL):
     global initialized, _MINIMAL_TYPE_TRANS_TYPES, \
-           _NORMAL_TYPE_TRANS_IN, _NORMAL_TYPE_TRANS_OUT
+           _NORMAL_TYPE_TRANS_IN, _NORMAL_TYPE_TRANS_OUT, \
+           _guessTextualBlobEncodingWhenUsingFB20AndEarlier
 
     if initialized:
         raise ProgrammingError('Cannot initialize module more than once.')
@@ -277,6 +290,8 @@ def init(type_conv=1, concurrency_level=_k.DEFAULT_CONCURRENCY_LEVEL):
         _look_up_array_descriptor,
         _look_up_array_subtype,
         _Cursor_execute_exception_type_filter,
+        _validateTPB,
+        _trans_info,
       )
 
     globalz = globals()
@@ -288,15 +303,36 @@ def init(type_conv=1, concurrency_level=_k.DEFAULT_CONCURRENCY_LEVEL):
               1: 'typeconv_backcompat', # the default
             100: 'typeconv_23plus',
             199: 'typeconv_23plus_lowmem',
-            200: 'typeconv_24plus',     # considered the "ideal" as of KIDB 3.2
+            200: 'typeconv_24plus',     # considered "ideal" for KIDB 3.2
+
+            # Code 300 is considered "ideal" for KIDB 3.3 and FB 2.1+.  It
+            # causes textual blobs to be handled the same way as other textual
+            # types, so unicode encoding/decoding is performed automagically.
+            # When converting in the input direction, this doesn't work with
+            # any version of FB prior to FB 2.1, because the API doesn't make
+            # the blob's character set ID available.
+            # Note that we use the same typeConvModule for code 300 as for
+            # code 200; the textual-blob-related changes are implemented by
+            # adding 'BLOB' entries to _NORMAL_TYPE_TRANS_*.
+            300: 'typeconv_24plus',
           }
         chosenTypeConvModuleName = typeConvOptions[type_conv]
         typeConvModule = __import__('kinterbasdb.' + chosenTypeConvModuleName,
             globalz, locals(), (chosenTypeConvModuleName,)
           )
+
+        extraMinTypeTransEntries = [] # 2007.02.10
+
         if type_conv > 1:
-            _MINIMAL_TYPE_TRANS_TYPES = \
-              _MINIMAL_TYPE_TRANS_TYPES + ('TEXT_UNICODE',)
+            extraMinTypeTransEntries.append('TEXT_UNICODE')
+        if type_conv >= 300:
+            _guessTextualBlobEncodingWhenUsingFB20AndEarlier = True
+            extraMinTypeTransEntries.append('BLOB')
+
+        if extraMinTypeTransEntries:
+            _MINIMAL_TYPE_TRANS_TYPES = _MINIMAL_TYPE_TRANS_TYPES \
+              + tuple(extraMinTypeTransEntries)
+
 
     for name in BASELINE_TYPE_TRANSLATION_FACILITIES:
         if not name.startswith('OPT:'):
@@ -340,6 +376,13 @@ def init(type_conv=1, concurrency_level=_k.DEFAULT_CONCURRENCY_LEVEL):
     if type_conv > 1:
         _NORMAL_TYPE_TRANS_IN['TEXT_UNICODE'] = unicode_conv_in
         _NORMAL_TYPE_TRANS_OUT['TEXT_UNICODE'] = unicode_conv_out
+    if type_conv >= 300: # 2007.02.10
+        textBlobsAsTextConfig = {
+            'mode': 'materialize',
+            'treat_subtype_text_as_text': True
+          }
+        _NORMAL_TYPE_TRANS_IN['BLOB'] = textBlobsAsTextConfig
+        _NORMAL_TYPE_TRANS_OUT['BLOB'] = textBlobsAsTextConfig
 
     initialized = True
 
@@ -452,6 +495,11 @@ def create_database(*args):
     # DATABASE" variant of isc_dsql_execute_immediate, for those who care), see
     # Connection.execute_immediate.
 
+    # 2007.04.05: If the path specified in this CREATE DATABASE statement needs
+    # to be Unicode, we should accept that.
+    if len(args) >= 1 and isinstance(args[0], unicode):
+        args = (args[0].encode(_FS_ENCODING),) + args[1:]
+
     C_con = _k.create_database(*args)
     return Connection(_CConnection=C_con)
 
@@ -483,9 +531,10 @@ def raw_byte_to_int(raw_byte):
 # BlobReader, PreparedStatement and Cursor can't be instantiated from Python,
 # but are exposed here to support isinstance(o, kinterbasdb.Class) and the
 # like.
+Transaction = _k.Transaction # 2007.01
 BlobReader = _k.BlobReader
-PreparedStatement = _k.PreparedStatement
 Cursor = _k.Cursor
+PreparedStatement = _k.PreparedStatement
 
 if _EVENT_HANDLING_SUPPORTED:
     EventConduit = _k.EventConduit
@@ -495,35 +544,9 @@ class Connection(object):
       Represents a connection between the database client (the Python process)
     and the database server.
 
-      The basic behavior of this class is documented by the Python DB API
-    Specification 2.0; this docstring covers only some extensions.  Also see
-    the KInterbasDB Usage Guide (docs/usage.html).
-
-    Attributes:
-    $dialect (read-only):
-       The Interbase SQL dialect of the connection.  One of:
-         1 for Interbase < 6.0
-         2 for "migration mode"
-         3 for Interbase >= 6.0 and Firebird (the default)
-
-    $precision_mode (read/write): (DEPRECATED)
-         precision_mode is deprecated in favor of dynamic type translation
-       (see the [set|get]_type_trans_[in|out] methods, and the Usage Guide).
-       ---
-         precision_mode 0 (the default) represents database fixed point values
-       (NUMERIC/DECIMAL fields) as Python floats, potentially losing precision.
-         precision_mode 1 represents database fixed point values as scaled
-       Python integers, preserving precision.
-         For more information, see the KInterbasDB Usage Guide.
-
-    $server_version (read-only):
-         The version string of the database server to which this connection
-       is connected.
-
-    $default_tpb (read/write):
-         The transaction parameter buffer (TPB) that will be used by default
-       for new transactions opened in the context of this connection.
-         For more information, see the KInterbasDB Usage Guide.
+      The basic functionality of this class is documented by the Python DB API
+    Specification 2.0, while the large amount of additional functionality is
+    documented by the KInterbasDB Usage Guide (docs/usage.html).
     """
 
     def __init__(self, *args, **keywords_args):
@@ -545,11 +568,17 @@ class Connection(object):
         # creating one via Connection_connect.  (The create_database function
         # uses this facility, for example.)
         if '_CConnection' in keywords_args:
-            self._C_con = keywords_args['_CConnection']
-            # Since we given a pre-existing CConnection instance rather than
-            # creating it ourselves, we need to explicitly give it a reference
-            # to its Python companion (self):
-            _k.Connection_python_wrapper_obj_set(self._C_con, self)
+            C_con = self._C_con = keywords_args['_CConnection']
+            assert C_con is not None
+            # Since we were given a pre-existing CConnection instance rather
+            # than creating it ourselves, we need to explicitly give it a
+            # reference to its Python companion (self), and allow it to
+            # establish a main_transaction:
+            _k.Connection_python_wrapper_obj_set(C_con, self)
+
+            # Nullify the private properties that we don't have in this case:
+            self._charset = None
+            self._C_con_params = None
         else:
             n_nonkeyword = len(args)
             n_keyword = len(keywords_args)
@@ -646,19 +675,7 @@ class Connection(object):
             kinterbasdb.isc_tpb_* constants.  See the Interbase API guide for
             these constants' meanings.
         """
-        if self.group is not None:
-            self.group.begin()
-            return
-
-        if tpb is None:
-            tpb = self.default_tpb
-        else:
-            tpb = _validateTPB(tpb)
-
-        if not isinstance(tpb, str):
-            tpb = tpb.render()
-
-        _k.Connection_begin(self._C_con, tpb)
+        return self._main_trans.begin(tpb=tpb)
 
 
     def prepare(self):
@@ -668,12 +685,7 @@ class Connection(object):
         it will be performed implicitly by commit() in a 2PC.
           See also the method ConnectionGroup.prepare.
         """
-        if self.group is not None:
-            self.group.prepare()
-            return
-
-        _k.Connection_prepare(self._C_con)
-
+        self._main_trans.prepare()
 
     def commit(self, retaining=False):
         """
@@ -689,11 +701,7 @@ class Connection(object):
             "open" across commits.
               See IB 6 API Guide pages 75 and 291 for more info.
         """
-        if self.group is not None:
-            self.group.commit(retaining=retaining)
-            return
-
-        _k.Connection_commit(self._C_con, retaining)
+        return self._main_trans.commit(retaining=retaining)
 
 
     def savepoint(self, name):
@@ -706,7 +714,7 @@ class Connection(object):
             ...
             con.rollback(savepoint='BEGINNING_OF_SOME_SUBTASK')
         """
-        self.execute_immediate('SAVEPOINT ' + name)
+        return self._main_trans.savepoint(name)
 
 
     def rollback(self, retaining=False, savepoint=None):
@@ -726,14 +734,9 @@ class Connection(object):
               If a savepoint name is supplied, only rolls back as far as that
             savepoint, rather than rolling back the entire transaction.
         """
-        if savepoint is None:
-            if self.group is not None:
-                self.group.rollback(retaining=retaining)
-                return
-
-            _k.Connection_rollback(self._C_con, retaining)
-        else:
-            self.execute_immediate('ROLLBACK TO ' + savepoint)
+        return self._main_trans.rollback(
+            retaining=retaining, savepoint=savepoint
+          )
 
 
     def execute_immediate(self, sql):
@@ -746,7 +749,17 @@ class Connection(object):
         create a cursor using the connection's cursor() method, then execute
         the statement using one of the cursor's execute methods.
         """
-        _k.Connection_execute_immediate(self._C_con, sql)
+        if isinstance(sql, unicode): # 2007.04.05
+            if self._charset:
+                from kinterbasdb import typeconv_text_unicode as ttu
+                pyCS = ttu.DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP[
+                    self._charset
+                  ]
+            else:
+                pyCS = 'ascii'
+            sql = sql.encode(pyCS)
+
+        return self._main_trans._execute_immediate(sql)
 
 
     def database_info(self, request, result_type):
@@ -961,52 +974,23 @@ class Connection(object):
 
 
     def transaction_info(self, request, result_type):
-        if not self._has_transaction():
-            raise ProgrammingError('This connection has no active transaction.')
-
-        return _k.Connection_transaction_info(self._C_con, request, result_type)
-
+        return self._main_trans.transaction_info(request, result_type)
 
     def trans_info(self, request):
-        # We process request as a sequence of info codes, even if only one code
-        # was supplied by the caller.
-        requestIsSingleton = isinstance(request, int)
-        if requestIsSingleton:
-            request = (request,)
+        return self._main_trans.trans_info(request)
 
-        results = {}
-        for infoCode in request:
-            if infoCode == globals().get('isc_info_tra_isolation', -1):
-                buf = self.transaction_info(infoCode, 's')
-                buf = buf[1 + struct.calcsize('h'):]
-                if len(buf) == 1:
-                    results[infoCode] = portable_int(buf)
-                else:
-                    # For isolation level isc_info_tra_read_committed, the
-                    # first byte indicates the isolation level
-                    # (isc_info_tra_read_committed), while the second indicates
-                    # the record version flag (isc_info_tra_rec_version or
-                    # isc_info_tra_no_rec_version).
-                    isolationLevelByte, recordVersionByte = struct.unpack('cc', buf)
-                    isolationLevel = portable_int(isolationLevelByte)
-                    recordVersion = portable_int(recordVersionByte)
-                    results[infoCode] = (isolationLevel, recordVersion)
-            else:
-                # At the time of this writing (2006.02.09),
-                # isc_info_tra_isolation is the only known return value of
-                # isc_transaction_info that's not a simple integer.
-                results[infoCode] = self.transaction_info(infoCode, 'i')
 
-        if requestIsSingleton:
-            return results[request[0]]
-        else:
-            return results
+    def trans(self, tpb=None):
+        "Creates a new Transaction that operates within the context of this"
+        " connection.  Cursors can be created within that Transaction via its"
+        " .cursor() method."
+        return Transaction(con=self, tpb=tpb)
 
 
     def cursor(self):
         "Creates a new cursor that operates within the context of this"
-        " connection."
-        return Cursor(self)
+        " connection's main_transaction."
+        return self._main_trans.cursor()
 
 
     def close(self):
@@ -1054,7 +1038,7 @@ class Connection(object):
     def _has_transaction(self):
         # Does this connection currently have an active transaction (including
         # a distributed transaction)?
-        return _k.Connection_has_transaction(self._C_con)
+        return _k.Connection_has_active_transaction(self._C_con)
 
     def _normalize_type_trans(self):
         # Set the type translation dictionaries to their "normal" form--the
@@ -1212,9 +1196,19 @@ class Connection(object):
     charset = property(_charset_get, _charset_set)
 
 
+    def _guessBlobCharSetIDFromConnectionCharSet(self): # 2007.02.10: HACK
+        if _guessTextualBlobEncodingWhenUsingFB20AndEarlier:
+            from kinterbasdb.typeconv_text_unicode \
+              import DB_CHAR_SET_NAME_TO_DB_CHAR_SET_ID_MAP
+            return DB_CHAR_SET_NAME_TO_DB_CHAR_SET_ID_MAP.get(
+                self._charset, None
+              )
+        else:
+            return None
+
     # group read-only property:
     def _group_get(self):
-        return _k.Connection_get_group(self._C_con)
+        return self._main_trans._group
     group = property(_group_get)
 
 
@@ -1222,17 +1216,7 @@ class Connection(object):
         # This package-private method allows ConnectionGroup's membership
         # management functionality to bypass the conceptually read-only nature
         # of the Connection.group property.
-        if group is not None:
-            assert _k.Connection_get_group(self._C_con) is None
-        # Unless the group is being cleared (set to None), pass a *WEAK*
-        # reference down to the C level (otherwise, even the cyclic garbage
-        # collector can't collect ConnectionGroups or their Connections because
-        # the Connections' references are held at the C level, not the Python
-        # level).
-        if group is None:
-            _k.Connection_set_group(self._C_con, None)
-        else:
-            _k.Connection_set_group(self._C_con, group)
+        self._main_trans._group = group
 
 
     def _ensure_group_membership(self, must_be_member, err_msg):
@@ -1250,11 +1234,55 @@ class Connection(object):
         return _k.Connection_timeout_enabled(self._C_con)
     _timeout_enabled = property(_timeout_enabled_get)
 
+    def _main_trans_get(self):
+        # NOTE:  We do not store a reference to the Transaction object within
+        # this Python Connection class, because that would defeat the
+        # Python-ref-cycle-free nature of the underlying C code.
+        return _k.Connection_main_trans_get(self._C_con)
+    _main_trans = property(_main_trans_get)
+
+    def _main_transaction_get(self):
+        # NOTE:  We do not store a reference to the
+        # ExternallyVisibleMainTransaction object within this Python Connection
+        # class, because that would defeat the Python-ref-cycle-free nature of
+        # the underlying C code.
+        return ExternallyVisibleMainTransaction(self._main_trans)
+    main_transaction = property(_main_transaction_get)
+
+    def _transactions_get(self):
+        return _k.Connection_transactions_get(self._C_con)
+    transactions = property(_transactions_get)
+
 
     def _activity_stamps(self):
         return _k.Connection__read_activity_stamps(self._C_con)
 
 
+class ExternallyVisibleMainTransaction(Transaction):
+    # Accept an internal (Transaction_is_main(t)) kinterbasdb.Transaction
+    # instance and wrap it, impersonating it as closely as possible.  Since
+    # the internal Transaction object does not own a reference to its
+    # Connection (*unlike* Transaction objects explicitly created by client
+    # code), this proxy ensures that the Connection remains alive at least as
+    # long as externally held references to its internal Transaction.
+    def __init__(self, rawMainTrans):
+        Transaction.__setattr__(self, '_rawMainTrans', rawMainTrans)
+        Transaction.__setattr__(self, '_rawMainTrans_con',
+            rawMainTrans.connection
+          )
+
+    def __getattribute__(self, name):
+        return getattr(
+            Transaction.__getattribute__(self, '_rawMainTrans'), name
+          )
+
+    def __cmp__(self, other):
+        return cmp(Transaction.__getattribute__(self, '_rawMainTrans'), other)
+
+    def __hash__(self):
+        return hash(Transaction.__getattribute__(self, '_rawMainTrans'))
+
+
 class ConnectionGroup(object):
     # XXX: ConnectionGroup objects currently are not thread-safe.  Since
     # separate Connections can be manipulated simultaneously by different
@@ -1349,9 +1377,7 @@ class ConnectionGroup(object):
     def remove(self, con):
         if con not in self:
             raise ProgrammingError('con is not a member of this group.')
-        # The following assertion was invalidated by the introduction of weak
-        # refs:
-        #assert con.group is self
+        assert con.group is self
         self._require_transaction_state(False,
             'Cannot remove connection from group that has an unresolved'
             ' transaction.'
@@ -1409,7 +1435,7 @@ class ConnectionGroup(object):
             'Must resolve current transaction before starting another.'
           )
         self._require_non_empty_group('start')
-        self._trans_handle = _k.distributed_begin(self._cons)
+        self._trans_handle = _k.distributed_begin(self, self._cons)
 
 
     def prepare(self):
@@ -1434,12 +1460,9 @@ class ConnectionGroup(object):
         if self._trans_handle is None:
             return
 
-        _k.distributed_commit(self._trans_handle, retaining)
+        _k.distributed_commit(self, self._trans_handle, self._cons, retaining)
         self._trans_handle = None
 
-        for con in self._cons: # 2005.07.22
-            _k.Connection_clear_transaction_stats(con._C_con)
-
 
     def rollback(self, retaining=False):
         self._require_non_empty_group('roll back')
@@ -1449,12 +1472,9 @@ class ConnectionGroup(object):
         if self._trans_handle is None:
             return
 
-        _k.distributed_rollback(self._trans_handle, retaining)
+        _k.distributed_rollback(self, self._trans_handle, self._cons, retaining)
         self._trans_handle = None
 
-        for con in self._cons: # 2005.07.22
-            _k.Connection_clear_transaction_stats(con._C_con)
-
 
 ##########################################
 ##     PUBLIC CLASSES: END              ##
@@ -1593,7 +1613,7 @@ class _DPBBuilder(object):
         self.addStringIfProvided(isc_dpb_password, password)
         self.addStringIfProvided(isc_dpb_sql_role_name, role)
 
-        self.charset = (charset and charset) or None
+        self.charset = (charset and charset.upper()) or None
         if self.charset:
             self.addString(isc_dpb_lc_ctype, charset)
 
@@ -1634,6 +1654,9 @@ class _DPBBuilder(object):
             else:
                 dsn = database
 
+        if _FS_ENCODING: # 2007.03.19
+            dsn = dsn.encode(_FS_ENCODING)
+
         assert dsn, 'Internal error in _build_connect_structures DSN prep.'
         return dsn
 
@@ -1807,19 +1830,22 @@ class _DPBBuilder(object):
   # isc_dpb_gstat_attach
   # isc_dpb_set_db_charset : s
 
-_DPB_CODES_WITH_STRING_VALUE = (
+
+_DPB_CODES_WITH_STRING_VALUE = [
     isc_dpb_user_name,
     isc_dpb_password,
     isc_dpb_lc_messages,
-    isc_dpb_set_db_charset,
-  )
+  ]
+if 'isc_dpb_set_db_charset' in globals():
+    _DPB_CODES_WITH_STRING_VALUE.append(isc_dpb_set_db_charset)
 
-_DPB_CODE_WITH_INT_VALUE = (
+
+_DPB_CODE_WITH_INT_VALUE = [
     isc_dpb_connect_timeout,
     isc_dpb_dummy_packet_interval,
     isc_dpb_sql_dialect,
     isc_dpb_set_db_sql_dialect,
-  )
+  ]
 
 
 def _trans_require_dict(obj):
@@ -2057,11 +2083,11 @@ class TableReservation(object):
         for tableName, resDefs in self.iteritems():
             tableNameLenWithTerm = len(tableName) + 1
             for (sharingMode, accessMode) in resDefs:
-                _(sharingMode)
                 _(accessMode)
                 _(struct.pack('<b%ds' % tableNameLenWithTerm,
                     tableNameLenWithTerm, tableName
                   ))
+                _(sharingMode)
 
         return ''.join(frags)
 
@@ -2351,3 +2377,40 @@ _TRANSACTION_CONFLICT_RAW_CODES = (
     # lock_timeout:  lock time-out on wait transaction
     #--------------------------------------------------------------------------
   )
+
+def _trans_info(trans, request):
+    # We process request as a sequence of info codes, even if only one code
+    # was supplied by the caller.
+    requestIsSingleton = isinstance(request, int)
+    if requestIsSingleton:
+        request = (request,)
+
+    results = {}
+    for infoCode in request:
+        # The global().get(...) workaround is here because only recent versions
+        # of FB expose constant isc_info_tra_isolation:
+        if infoCode == globals().get('isc_info_tra_isolation', -1):
+            buf = trans.transaction_info(infoCode, 's')
+            buf = buf[1 + struct.calcsize('h'):]
+            if len(buf) == 1:
+                results[infoCode] = portable_int(buf)
+            else:
+                # For isolation level isc_info_tra_read_committed, the
+                # first byte indicates the isolation level
+                # (isc_info_tra_read_committed), while the second indicates
+                # the record version flag (isc_info_tra_rec_version or
+                # isc_info_tra_no_rec_version).
+                isolationLevelByte, recordVersionByte = struct.unpack('cc', buf)
+                isolationLevel = portable_int(isolationLevelByte)
+                recordVersion = portable_int(recordVersionByte)
+                results[infoCode] = (isolationLevel, recordVersion)
+        else:
+            # At the time of this writing (2006.02.09),
+            # isc_info_tra_isolation is the only known return value of
+            # isc_transaction_info that's not a simple integer.
+            results[infoCode] = trans.transaction_info(infoCode, 'i')
+
+    if requestIsSingleton:
+        return results[request[0]]
+    else:
+        return results
