@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Parameter Conversion
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -99,7 +99,7 @@ static InputStatus PyObject2XSQLVAR(
   const short data_type = XSQLVAR_SQLTYPE_IGNORING_NULL_FLAG(sqlvar);
   const short data_subtype = sqlvar->sqlsubtype;
   const short scale = sqlvar->sqlscale;
-  const unsigned short dialect = cur->connection->dialect;
+  const unsigned short dialect = Transaction_get_dialect(cur->trans);
 
   boolean is_nonstandard_blob = FALSE;
 
@@ -136,6 +136,7 @@ static InputStatus PyObject2XSQLVAR(
 
     is_nonstandard_blob = (boolean)
       (data_type == SQL_BLOB && PyDict_Check(translator));
+
     if (!is_nonstandard_blob) {
       py_input_converted = dynamically_type_convert_input_obj_if_necessary(
           py_input,
@@ -168,6 +169,12 @@ static InputStatus PyObject2XSQLVAR(
         PyObject *py_blob_charset_id;
         boolean is_unicode_charset;
 
+        /* 2007.02.10: */
+        /* It's a textual blob that we've been ordered to treat like normal
+         * text, so we revert the is_nonstandard_blob flag to FALSE so that the
+         * converted input will be passed to the normal blob handler. */
+        is_nonstandard_blob = FALSE;
+
         if (get_blob_converter_override_for_direction(TRUE, cur, sqlvar,
                 &py_converter_override, &py_blob_charset_id, &is_unicode_charset
               ) != 0
@@ -192,20 +199,25 @@ static InputStatus PyObject2XSQLVAR(
             /* Pass 2-tuple of (py_input, py_blob_charset_id): */
             PyObject *tuple_of_unicode_obj_and_charset_code = PyTuple_New(2);
             if (tuple_of_unicode_obj_and_charset_code == NULL) { goto fail; }
+
             Py_INCREF(py_input);
             PyTuple_SET_ITEM(tuple_of_unicode_obj_and_charset_code, 0, py_input);
+
             Py_INCREF(py_blob_charset_id);
             PyTuple_SET_ITEM(tuple_of_unicode_obj_and_charset_code, 1,
                 py_blob_charset_id
               );
 
             py_input_converted = PyObject_CallFunctionObjArgs(
-                py_converter_override, tuple_of_unicode_obj_and_charset_code, NULL
+                py_converter_override, tuple_of_unicode_obj_and_charset_code,
+                NULL
               );
 
             Py_DECREF(tuple_of_unicode_obj_and_charset_code);
           }
         }
+        assert (py_blob_charset_id != NULL);
+        Py_DECREF(py_blob_charset_id);
       } else {
         /* Either it's not a textual blob, or it is a textual blob that we've
          * been ordered not to auto-[enc|de]code.
@@ -215,8 +227,6 @@ static InputStatus PyObject2XSQLVAR(
          * functions the same way naked translation would (i.e., it passes
          * the input string/buffer through unchanged). */
 
-        /* Wouldn't be here if treat_subtype_text_as_text were true: */
-        assert (!treat_subtype_text_as_text);
         if (mode == blob_mode_materialize) {
           /* For now, materialized mode with treat_subtype_text_as_text FALSE
            * behaves the same way as "conventional" naked translation. */
@@ -311,13 +321,11 @@ static InputStatus PyObject2XSQLVAR(
     break;
   #endif /* INTERBASE_6_OR_LATER */
 
-  #ifdef INTERBASE_7_OR_LATER
   case SQL_BOOLEAN:
     PyObject2XSQLVAR_TRY_INPUT_CONVERSION(
         conv_in_boolean_conventional(py_input_converted, sqlvar)
       );
     break;
-  #endif /* INTERBASE_7_OR_LATER */
 
   case SQL_BLOB:
     /* 2005.06.19: Added support for non-materialized blob input. */
@@ -336,8 +344,8 @@ static InputStatus PyObject2XSQLVAR(
           conv_in_blob_from_pyfilelike(
               py_input_converted,
               (ISC_QUAD *) sqlvar->sqldata,
-              cur->status_vector, cur->connection->db_handle,
-              CON_GET_TRANS_HANDLE(cur->connection) /* 2003.10.15a:OK */
+              cur->status_vector, *Transaction_get_db_handle_p(cur->trans),
+              *Transaction_get_handle_p(cur->trans)
             )
         );
     }
@@ -711,7 +719,7 @@ PyObject *XSQLDA2Description(XSQLDA *sqlda, Cursor *cur) {
      * instead of returning the same type regardless. */
     if (data_type != SQL_ARRAY) {
       PyObject *translator_key = _get_cached_type_name_for_conventional_code(
-          cur->connection->dialect,
+          Transaction_get_dialect(cur->trans),
           data_type, sqlvar->sqlsubtype, sqlvar->sqlscale
         );
       if (translator_key == NULL) { goto fail; }
@@ -789,12 +797,10 @@ PyObject *XSQLDA2Description(XSQLDA *sqlda, Cursor *cur) {
       break;
     #endif /* INTERBASE_6_OR_LATER */
 
-    #ifdef INTERBASE_7_OR_LATER
     case SQL_BOOLEAN:
       DEFAULT_TYPE_IS(PyBool_Type);
       display_size = 5;
       break;
-    #endif /* INTERBASE_7_OR_LATER */
 
     case SQL_ARRAY:
       DEFAULT_TYPE_IS(PyList_Type);
@@ -894,10 +900,10 @@ static PyObject *XSQLVAR2PyObject(
 
   const short data_type = XSQLVAR_SQLTYPE_IGNORING_NULL_FLAG(sqlvar);
   const short data_subtype = sqlvar->sqlsubtype;
-  const unsigned short dialect = cur->connection->dialect;
+  const unsigned short dialect = Transaction_get_dialect(cur->trans);
 
   const boolean is_array = (boolean) (data_type == SQL_ARRAY);
-  boolean is_nonstandard_blob = FALSE;
+  boolean is_nonstandard_blob;
 
   /* Array DTT is activated elsewhere--see kiconversion_array.c */
   if (!is_array) {
@@ -909,6 +915,13 @@ static PyObject *XSQLVAR2PyObject(
     if (converter == NULL) { goto fail; }
   }
 
+  /* 2007.02.10: */
+  /* The determination of is_nonstandard_blob needs to be made before we check
+   * for SQL NULL below, so that if SQL NULL is found and we jump to 'succeed',
+   * the 'succeed' clause won't mistakenly try to use a standard converter for
+   * a nonstandard blob: */
+  is_nonstandard_blob = (data_type == SQL_BLOB && PyDict_Check(converter));
+
   if (   XSQLVAR_IS_ALLOWED_TO_BE_NULL(sqlvar)
       && XSQLVAR_IS_NULL(sqlvar)
      )
@@ -976,11 +989,9 @@ static PyObject *XSQLVAR2PyObject(
     break;
   #endif /* INTERBASE_6_OR_LATER */
 
-  #ifdef INTERBASE_7_OR_LATER
   case SQL_BOOLEAN:
     result = conv_out_boolean(sqlvar->sqldata);
     break;
-  #endif /* INTERBASE_7_OR_LATER */
 
   case SQL_BLOB:
     /* 2005.06.21: */
@@ -996,12 +1007,11 @@ static PyObject *XSQLVAR2PyObject(
     {
     ISC_QUAD *blob_id = (ISC_QUAD *) sqlvar->sqldata;
 
-    is_nonstandard_blob = (boolean) PyDict_Check(converter);
     if (!is_nonstandard_blob) {
       /* Backward-compatible behavior (materialized, no auto-decoding): */
       result = conv_out_blob_materialized(blob_id, cur->status_vector,
-          cur->connection->db_handle,
-          CON_GET_TRANS_HANDLE(cur->connection) /* 2003.10.15a:OK */
+          *Transaction_get_db_handle_p(cur->trans),
+          *Transaction_get_handle_p(cur->trans)
         );
     } else {
       BlobMode mode;
@@ -1030,16 +1040,14 @@ static PyObject *XSQLVAR2PyObject(
 
       if (mode == blob_mode_materialize) {
         result = conv_out_blob_materialized(blob_id, cur->status_vector,
-            cur->connection->db_handle,
-            CON_GET_TRANS_HANDLE(cur->connection) /* 2003.10.15a:OK */
+            *Transaction_get_db_handle_p(cur->trans),
+            *Transaction_get_handle_p(cur->trans)
           );
         if (treat_subtype_text_as_text__applies) {
           /* Look up the character set ID of this blob to determine whether to
            * route it through the TEXT or TEXT_UNICODE dynamic type translator
            * for auto-decoding. */
 
-          /* Both py_converter_override and py_blob_charset_id will hold only
-           * borrowed refs, so no worries about DECREFing them. */
           PyObject *py_converter_override;
           PyObject *py_blob_charset_id;
           boolean is_unicode_charset;
@@ -1066,11 +1074,9 @@ static PyObject *XSQLVAR2PyObject(
               /* Pass 2-tuple of (result, py_blob_charset_id): */
               PyObject *tuple_of_raw_string_and_charset_code = PyTuple_New(2);
               if (tuple_of_raw_string_and_charset_code == NULL) { goto fail; }
-              /* We pass reference ownership over result to the tuple, and since
-               * we currently hold only a borrowed reference to
-               * py_blob_charset_id, we create an artifical ref to be passed to
-               * the tuple. */
+              /* We pass reference ownership over result to the tuple. */
               PyTuple_SET_ITEM(tuple_of_raw_string_and_charset_code, 0, result);
+
               Py_INCREF(py_blob_charset_id);
               PyTuple_SET_ITEM(tuple_of_raw_string_and_charset_code, 1,
                   py_blob_charset_id
@@ -1085,9 +1091,11 @@ static PyObject *XSQLVAR2PyObject(
                * NULL result and indicate an error if necessary. */
             }
           }
+          assert (py_blob_charset_id != NULL);
+          Py_DECREF(py_blob_charset_id);
         }
       } else {
-        result = (PyObject *) BlobReader_create(cur->connection);
+        result = (PyObject *) BlobReader_create(cur->trans);
         if (result == NULL) { goto fail; }
         if (BlobReader_open((BlobReader *) result, blob_id) != 0) { goto fail; }
         assert (BlobReader_is_open((BlobReader *) result));
@@ -1103,8 +1111,8 @@ static PyObject *XSQLVAR2PyObject(
           (ISC_QUAD *) sqlvar->sqldata,
 
           cur->status_vector,
-          &cur->connection->db_handle,
-          CON_GET_TRANS_HANDLE_ADDR(cur->connection), /* 2003.10.15a:OK */
+          Transaction_get_db_handle_p(cur->trans),
+          Transaction_get_handle_p(cur->trans),
 
           sqlvar->relname, sqlvar->relname_length,
           sqlvar->sqlname, sqlvar->sqlname_length
@@ -1130,6 +1138,7 @@ static PyObject *XSQLVAR2PyObject(
   succeed:
     /* Nonstandard blobs and arrays are subject to special dynamic type
      * translation, which is not applied in the standard manner. */
+    assert (result != NULL);
     if (!(is_nonstandard_blob || is_array)) {
       assert (converter != NULL); /* Can't be NULL, but may be None. */
       /* Replacing the PyObject pointer in result is *not* a refcount leak; see
@@ -1154,15 +1163,20 @@ PyObject *XSQLDA2Tuple(Cursor *cur, XSQLDA *sqlda) {
   PyObject *record = PyTuple_New(sqlvar_count);
   if (record == NULL) { return NULL; }
 
-  for (sqlvar_index = 0; sqlvar_index < sqlvar_count; sqlvar_index++) {
+  for (sqlvar_index = 0; sqlvar_index < sqlvar_count; ++sqlvar_index) {
     var = XSQLVAR2PyObject(cur, sqlvar_index, sqlda->sqlvar + sqlvar_index);
     if (var == NULL) {
-      /* XSQLVAR2PyObject will have set an exception; we can just return NULL. */
-      Py_DECREF(record);
-      return NULL;
+      /* XSQLVAR2PyObject will have set an exception. */
+      goto fail;
     }
+    /* PyTuple_SET_ITEM steals our ref to var. */
     PyTuple_SET_ITEM(record, sqlvar_index, var);
   }
 
   return record;
+
+  fail:
+    assert (PyErr_Occurred());
+    Py_XDECREF(record);
+    return NULL;
 } /* XSQLDA2Tuple */
