@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Array Conversion (both ways)
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -72,8 +72,10 @@ static PyObject *conv_out_array(
     Cursor *cursor, short sqlvar_index,
     ISC_QUAD *array_id,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
-    char *rel_name, short rel_name_length, char *field_name, short field_name_length
+    ISC_STATUS *status_vector,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle_p,
+    char *rel_name, short rel_name_length,
+    char *field_name, short field_name_length
   );
 
 static PyObject *conv_out_array_element(
@@ -83,8 +85,10 @@ static PyObject *conv_out_array_element(
     short data_type, size_t size_of_single_element,
     short scale,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
-    char *rel_name, short rel_name_length, char *field_name, short field_name_length
+    ISC_STATUS *status_vector,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    char *rel_name, short rel_name_length,
+    char *field_name, short field_name_length
   );
 
 static PyObject *_extract_db_array_buffer_to_pyseq(
@@ -95,8 +99,10 @@ static PyObject *_extract_db_array_buffer_to_pyseq(
     short data_type, size_t size_of_single_element,
     short scale,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
-    char *rel_name, short rel_name_length, char *field_name, short field_name_length
+    ISC_STATUS *status_vector, isc_db_handle *db_handle,
+    isc_tr_handle *trans_handle_p,
+    char *rel_name, short rel_name_length,
+    char *field_name, short field_name_length
   );
 
 
@@ -104,7 +110,8 @@ static PyObject *_extract_db_array_buffer_to_pyseq(
 static InputStatus conv_in_array(
     PyObject *py_input, ISC_QUAD **array_id_slot,
     Cursor *cursor, short sqlvar_index,
-    char *rel_name, short rel_name_length, char *field_name, short field_name_length
+    char *rel_name, short rel_name_length,
+    char *field_name, short field_name_length
   );
 
 static InputStatus conv_in_array_element(
@@ -115,7 +122,8 @@ static InputStatus conv_in_array_element(
     size_t size_of_single_element, short scale,
     PyObject *converter,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    ISC_STATUS *status_vector, isc_db_handle *db_handle,
+    isc_tr_handle *trans_handle_p,
     Cursor *cur
   );
 
@@ -130,14 +138,15 @@ static InputStatus _extract_pyseq_to_db_array_buffer(
     size_t size_of_single_element, short scale,
     PyObject *converter,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    ISC_STATUS *status_vector, isc_db_handle *db_handle,
+    isc_tr_handle *trans_handle_p,
     Cursor *cur
   );
 
 
 /* Functions common to both input and output: */
 static ISC_ARRAY_DESC *_look_up_array_descriptor(
-    CConnection *con,
+    Transaction *trans,
     /* These strings aren't null-terminated: */
     char *sqlvar_rel_name, short sqlvar_rel_name_length,
     char *sqlvar_field_name, short sqlvar_field_name_length
@@ -150,7 +159,7 @@ static short *_extract_dimensions_sizes(
   );
 
 static short _determine_sqlsubtype_for_array(
-    CConnection *con,
+    Transaction *trans,
     char *rel_name, short rel_name_length,
     char *field_name, short field_name_length
   );
@@ -181,15 +190,14 @@ static InputStatus conv_in_array(
   size_t source_buf_size;
 
   ISC_STATUS *status_vector = cursor->status_vector;
-  isc_db_handle *db_handle = &cursor->connection->db_handle;
-  isc_tr_handle *trans_handle = CON_GET_TRANS_HANDLE_ADDR(cursor->connection); /* 2003.10.15a:OK */
+  isc_db_handle *db_handle = Transaction_get_db_handle_p(cursor->trans);
+  isc_tr_handle *trans_handle_p = Transaction_get_handle_p(cursor->trans);
 
   PyObject *converter = NULL;
   short data_subtype = -1;
 
   /* Read the database array descriptor for this field. */
-  desc = _look_up_array_descriptor(
-      cursor->connection,
+  desc = _look_up_array_descriptor(cursor->trans,
       rel_name, rel_name_length,
       field_name, field_name_length
     );
@@ -218,11 +226,11 @@ static InputStatus conv_in_array(
   source_buf_walker = source_buf;
 
   assert (data_type != -1);
-  data_subtype = _determine_sqlsubtype_for_array(cursor->connection,
+  data_subtype = _determine_sqlsubtype_for_array(cursor->trans,
       rel_name, rel_name_length, field_name, field_name_length
     );
   if (data_subtype == SQLSUBTYPE_DETERMINATION_ERROR) { goto fail; }
-  ++cursor->connection->n_prepared_statements_executed_since_trans_start;
+  Transaction_stats_record_ps_executed(cursor->trans);
 
   { short scale = desc->array_desc_scale;
 
@@ -243,12 +251,12 @@ static InputStatus conv_in_array(
         /* For conversion: */
         &source_buf_walker,
 
-        cursor->connection->dialect,
+        Transaction_get_dialect(cursor->trans),
         data_type, data_subtype,
         size_of_single_element, scale,
         converter,
 
-        status_vector, db_handle, trans_handle,
+        status_vector, db_handle, trans_handle_p,
         cursor
       );
     if (status != INPUT_OK) { goto fail; }
@@ -275,7 +283,7 @@ static InputStatus conv_in_array(
 
     ENTER_GDAL
     isc_array_put_slice(status_vector,
-        db_handle, trans_handle,
+        db_handle, trans_handle_p,
         array_id, desc, source_buf, (ISC_LONG *) &source_buf_size
       );
     LEAVE_GDAL
@@ -341,7 +349,8 @@ static InputStatus conv_in_array_element(
     size_t size_of_single_element, short scale,
     PyObject *converter,
 
-    ISC_STATUS *status_vector, isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    ISC_STATUS *status_vector, isc_db_handle *db_handle,
+    isc_tr_handle *trans_handle_p,
     Cursor *cur
   )
 {
@@ -427,13 +436,11 @@ static InputStatus conv_in_array_element(
     break;
   #endif /* INTERBASE_6_OR_LATER */
 
-  #ifdef INTERBASE_7_OR_LATER
   case blr_boolean_dtype:
     CONV_IN_ARRAY_ELEMENT_TRY_INPUT_CONVERSION(
         conv_in_boolean_array(py_input_converted, data_slot)
       );
     break;
-  #endif /* INTERBASE_7_OR_LATER */
 
   /* Currently, none of the following types is supported, because it's not
    * clear how one would create such a field via SQL DDL.  As far as I can
@@ -512,7 +519,7 @@ static InputStatus _extract_pyseq_to_db_array_buffer(
     PyObject *converter,
 
     ISC_STATUS *status_vector,
-    isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle_p,
     Cursor *cur
   )
 {
@@ -603,7 +610,7 @@ static InputStatus _extract_pyseq_to_db_array_buffer(
             size_of_single_element, scale,
             converter,
 
-            status_vector, db_handle, trans_handle,
+            status_vector, db_handle, trans_handle_p,
             cur
           );
 
@@ -636,7 +643,7 @@ static InputStatus _extract_pyseq_to_db_array_buffer(
             size_of_single_element, scale,
             converter,
 
-            status_vector, db_handle, trans_handle,
+            status_vector, db_handle, trans_handle_p,
             cur
           );
 
@@ -672,7 +679,7 @@ static PyObject *conv_out_array(
     Cursor *cursor, short sqlvar_index,
     ISC_QUAD *array_id,
     ISC_STATUS *status_vector,
-    isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle_p,
     char *rel_name, short rel_name_length,
     char *field_name, short field_name_length
   )
@@ -693,8 +700,7 @@ static PyObject *conv_out_array(
 
   int total_number_of_elements; /* Will be set later. */
 
-  ISC_ARRAY_DESC *desc = _look_up_array_descriptor(
-      cursor->connection,
+  ISC_ARRAY_DESC *desc = _look_up_array_descriptor(cursor->trans,
       rel_name, rel_name_length,
       field_name, field_name_length
     );
@@ -726,7 +732,7 @@ static PyObject *conv_out_array(
 
   ENTER_GDAL
   isc_array_get_slice(status_vector,
-      db_handle, trans_handle,
+      db_handle, trans_handle_p,
       array_id, desc,
       (void *) output_buf, (ISC_LONG *) &output_buf_size
     );
@@ -745,7 +751,7 @@ static PyObject *conv_out_array(
       /* Boilerplate parameters: */
       data_type, size_of_single_element, scale,
       /* Pass through utility stuff from above: */
-      status_vector, db_handle, trans_handle,
+      status_vector, db_handle, trans_handle_p,
       rel_name, rel_name_length, field_name, field_name_length
     );
   if (result == NULL) { goto fail; }
@@ -790,7 +796,7 @@ static PyObject *_extract_db_array_buffer_to_pyseq(
     short data_type, size_t size_of_single_element,
     short scale,
     ISC_STATUS *status_vector,
-    isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle_p,
     char *rel_name, short rel_name_length,
     char *field_name, short field_name_length
   )
@@ -811,7 +817,7 @@ static PyObject *_extract_db_array_buffer_to_pyseq(
 
           data_type, size_of_single_element,
           scale,
-          status_vector, db_handle, trans_handle,
+          status_vector, db_handle, trans_handle_p,
           rel_name, rel_name_length, field_name, field_name_length
         );
       if (val == NULL) { goto fail; }
@@ -832,7 +838,7 @@ static PyObject *_extract_db_array_buffer_to_pyseq(
 
           /* Boilerplate parameters (capitalized to differentiate them): */
           data_type, size_of_single_element, scale,
-          status_vector, db_handle, trans_handle,
+          status_vector, db_handle, trans_handle_p,
           rel_name, rel_name_length, field_name, field_name_length
         );
       if (subList == NULL) { goto fail; }
@@ -863,21 +869,20 @@ static PyObject *conv_out_array_element(
     short scale,
 
     ISC_STATUS *status_vector,
-    isc_db_handle *db_handle, isc_tr_handle *trans_handle,
+    isc_db_handle *db_handle, isc_tr_handle *trans_handle_p,
     char *rel_name, short rel_name_length,
     char *field_name, short field_name_length
   )
 {
   PyObject *result = NULL;
   PyObject *converter = NULL;
-  const unsigned short dialect = cursor->connection->dialect;
+  const unsigned short dialect = Transaction_get_dialect(cursor->trans);
 
-  const short data_subtype = _determine_sqlsubtype_for_array(
-      cursor->connection,
+  const short data_subtype = _determine_sqlsubtype_for_array(cursor->trans,
       rel_name, rel_name_length, field_name, field_name_length
     );
   if (data_subtype == SQLSUBTYPE_DETERMINATION_ERROR) { goto fail; }
-  ++cursor->connection->n_prepared_statements_executed_since_trans_start;
+  Transaction_stats_record_ps_executed(cursor->trans);
 
   converter = cursor_get_out_converter(cursor, sqlvar_index,
       data_type, data_subtype, scale, TRUE
@@ -962,11 +967,9 @@ static PyObject *conv_out_array_element(
     break;
   #endif /* INTERBASE_6_OR_LATER */
 
-  #ifdef INTERBASE_7_OR_LATER
   case blr_boolean_dtype:
     result = conv_out_boolean(data);
     break;
-  #endif /* INTERBASE_7_OR_LATER */
 
   case blr_quad:
     /* ISC_QUAD structure; since the DB engine doesn't support arrays of
@@ -1017,15 +1020,15 @@ static PyObject *conv_out_array_element(
 
 #ifdef USE_OFFICIAL_ISC_ARRAY_LOOKUP_BOUNDS
 static ISC_ARRAY_DESC *_look_up_array_descriptor(
-    CConnection *con,
+    Transaction *trans,
     /* These strings aren't null-terminated: */
     char *sqlvar_rel_name, short sqlvar_rel_name_length,
     char *sqlvar_field_name, short sqlvar_field_name_length
   )
 {
-  ISC_STATUS *status_vector = con->status_vector;
-  isc_db_handle *db_handle = &con->db_handle;
-  isc_tr_handle *trans_handle = CON_GET_TRANS_HANDLE_ADDR(con); /* 2003.10.15a:OK */
+  ISC_STATUS *status_vector = Transaction_get_sv(trans);
+  isc_db_handle *db_handle = Transaction_get_db_handle_p(trans);
+  isc_tr_handle *trans_handle_p = Transaction_get_handle_p(trans);
 
   /* isc_array_lookup_* functions require null-terminated strings, but the
    * relevant strings from the XSQLVAR structure are not null-terminated. */
@@ -1051,7 +1054,7 @@ static ISC_ARRAY_DESC *_look_up_array_descriptor(
 
   ENTER_GDAL
   isc_array_lookup_bounds(status_vector,
-      db_handle, trans_handle,
+      db_handle, trans_handle_p,
       null_terminated_table_name, null_terminated_field_name,
       desc
     );
@@ -1088,7 +1091,7 @@ static ISC_ARRAY_DESC *_look_up_array_descriptor(
 } /* _look_up_array_descriptor */
 #else
 static ISC_ARRAY_DESC *_look_up_array_descriptor(
-    CConnection *con,
+    Transaction *trans,
     /* These strings aren't null-terminated: */
     char *sqlvar_rel_name, short sqlvar_rel_name_length,
     char *sqlvar_field_name, short sqlvar_field_name_length
@@ -1100,6 +1103,9 @@ static ISC_ARRAY_DESC *_look_up_array_descriptor(
   PyObject *py_rel_name = NULL;
   PyObject *py_field_name = NULL;
 
+  assert (trans != NULL);
+  assert (trans->con_python_wrapper != NULL);
+
   py_rel_name = PyString_FromStringAndSize(
       sqlvar_rel_name, sqlvar_rel_name_length
     );
@@ -1111,7 +1117,7 @@ static ISC_ARRAY_DESC *_look_up_array_descriptor(
   if (py_field_name == NULL) { goto fail; }
 
   py_desc = PyObject_CallFunctionObjArgs(py_look_up_array_descriptor,
-      con->python_wrapper_obj, py_rel_name, py_field_name, NULL
+      trans->con_python_wrapper, py_rel_name, py_field_name, NULL
     );
   if (py_desc == NULL) { goto fail; }
 
@@ -1183,15 +1189,15 @@ static short *_extract_dimensions_sizes(
 } /* _extract_dimensions_sizes */
 
 static short _determine_sqlsubtype_for_array(
-    CConnection *con,
+    Transaction *trans,
     char *rel_name, short rel_name_length,
     char *field_name, short field_name_length
   )
 {
   #ifdef USE_OFFICIAL_ISC_ARRAY_LOOKUP_BOUNDS
-    ISC_STATUS *status_vector = con->status_vector;
-    isc_db_handle *db_handle = &con->db_handle;
-    isc_tr_handle *trans_handle = CON_GET_TRANS_HANDLE_ADDR(con); /* 2003.10.15a:OK */
+    ISC_STATUS *status_vector = Transaction_get_sv(trans);
+    isc_db_handle *db_handle = Transaction_get_db_handle_p(trans);
+    isc_tr_handle *trans_handle_p = Transaction_get_handle_p(trans);
     short out_var_sqlind = SQLIND_NULL;
 
     /* Returns the subtype on success, or SQLSUBTYPE_DETERMINATION_ERROR on error. */
@@ -1274,7 +1280,7 @@ static short _determine_sqlsubtype_for_array(
 
     ENTER_GDAL
     isc_dsql_prepare(status_vector,
-        trans_handle, &stmt_handle_sqlsubtype,
+        trans_handle_p, &stmt_handle_sqlsubtype,
         0, (char *) subtype_determination_statement, 3,
         out_da
       );
@@ -1283,7 +1289,7 @@ static short _determine_sqlsubtype_for_array(
 
     ENTER_GDAL
     isc_dsql_execute2(status_vector,
-        trans_handle, &stmt_handle_sqlsubtype,
+        trans_handle_p, &stmt_handle_sqlsubtype,
         3, in_da, out_da
       );
     LEAVE_GDAL
@@ -1327,13 +1333,16 @@ static short _determine_sqlsubtype_for_array(
       }
 
       return sqlsubtype;
-  #else
+  #else /* !USE_OFFICIAL_ISC_ARRAY_LOOKUP_BOUNDS: */
     PyObject *py_sqlsubtype = NULL;
     short sqlsubtype = SQLSUBTYPE_DETERMINATION_ERROR;
 
     PyObject *py_rel_name = NULL;
     PyObject *py_field_name = NULL;
 
+    assert (trans != NULL);
+    assert (trans->con_python_wrapper != NULL);
+
     py_rel_name = PyString_FromStringAndSize(rel_name, rel_name_length);
     if (py_rel_name == NULL) { goto fail; }
 
@@ -1341,7 +1350,7 @@ static short _determine_sqlsubtype_for_array(
     if (py_field_name == NULL) { goto fail; }
 
     py_sqlsubtype = PyObject_CallFunctionObjArgs(py_look_up_array_subtype,
-        con->python_wrapper_obj, py_rel_name, py_field_name, NULL
+        trans->con_python_wrapper, py_rel_name, py_field_name, NULL
       );
     if (py_sqlsubtype == NULL) { goto fail; }
 
