@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Materialized Blob Conversion
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -197,7 +197,7 @@ static PyObject *conv_out_blob_materialized_in_single_chunk(
 {
   ISC_LONG bytes_read_so_far = 0;
   unsigned short bytes_actually_read;
-  int blob_stat;
+  ISC_STATUS blob_stat; /* 2007.02.10: FB 2.1 fix: int->ISC_STATUS */
 
   char *py_str_start_ptr;
   /* Create an empty PyStringObject large enough to hold the entire chunk. */
@@ -376,7 +376,7 @@ static int _blob_info_total_size_and_max_segment_size(
     length = (short) isc_vax_integer(ptr, sizeof(short));
     LEAVE_GDAL
 
-    ptr += 2;
+    ptr += sizeof(short);
 
     switch (item) {
       case isc_info_blob_total_length:
