@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Support for Conversion of Nonstandard Blobs
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -152,15 +152,15 @@ static int validate_nonstandard_blob_config_dict(PyObject *config,
   }
 
   /* Validate 'treat_subtype_text_as_text' (optional boolean; defaults to
-   * true). */
+   * false). */
   py_treat_subtype_text_as_text = PyDict_GetItem(config,
       blob_nonstandard__config_slot_name__treat_subtype_text_as_text
     );
   if (py_treat_subtype_text_as_text == NULL) {
-    /* The value was missing; default to false (since the engine doesn't
-     * provide the supporting values required to determine the character set of
-     * incoming blobs, we obviously don't want to enable that feature by
-     * default). */
+    /* The value was missing; default to false (since the engine (prior to FB
+     * 2.1) doesn't provide the supporting values required to determine the
+     * character set of incoming blobs, we obviously don't want to enable that
+     * feature by default). */
     *treat_subtype_text_as_text = FALSE;
   } else {
     const int py_treat_subtype_text_as_text__as_bool =
@@ -184,11 +184,12 @@ static PyObject *determine_blob_character_set_id(
     Cursor *cursor, XSQLVAR *sqlvar
   )
 {
-  /* Returns a *borrowed* reference to a Python int containing the character
-   * set ID on success; NULL on failure. */
+  /* Returns a new reference to a Python int containing the character set ID
+   * on success; NULL on failure. */
   #ifdef USE_OFFICIAL_ISC_BLOB_LOOKUP_DESC
   ISC_BLOB_DESC blob_desc;
   #endif
+  Transaction *trans;
   CConnection *con;
   boolean might_be_cached = TRUE;
   PyObject *py_reldict = NULL;
@@ -197,10 +198,24 @@ static PyObject *determine_blob_character_set_id(
   PyObject *py_char_set_id = NULL;
 
   assert (cursor != NULL);
-  assert (cursor->connection != NULL);
+  assert (cursor->trans != NULL);
   assert (sqlvar != NULL);
 
-  con = cursor->connection;
+  /* 2007.02.10: */
+  /* Firebird 2.1 will store the blob's character set ID in the
+   * sqlvar->sqlscale field, which will make this function's task trivial.
+   * The relevant FB changelog entry is:
+   *   2006-08-08 13:23  asfernandes
+   *     firebird2/src/: dsql/dsql.cpp (1.178), isql/isql.epp (1.196):
+   *     Put blob charset in XSQLVAR::sqlscale
+   */
+  if (sqlvar->sqlscale != 0) { /* For FB 2.1+ */
+    return PyInt_FromLong(sqlvar->sqlscale);
+  }
+
+  trans = cursor->trans;
+  con = Transaction_get_con(trans);
+  assert (con != NULL);
 
   /* The fields XSQLVAR.relname and XSQLVAR.sqlname are not null-terminated,
    * but isc_blob_lookup_desc expects null-terminated strings.  Apparently the
@@ -208,6 +223,26 @@ static PyObject *determine_blob_character_set_id(
    * C string constants for the relation name and field name.  Welcome to the
    * 1980s. */
 
+  if (sqlvar->relname_length == 0 && sqlvar->sqlname_length == 0) {
+    PyObject *py_char_set_id = PyObject_CallMethod(con->python_wrapper_obj,
+        "_guessBlobCharSetIDFromConnectionCharSet", NULL
+      );
+    if (py_char_set_id != NULL && py_char_set_id != Py_None) {
+      return py_char_set_id;
+    } else {
+      raise_exception(OperationalError, "At this stage of execution, the"
+          " database client library has not provided enough information to"
+          " allow KInterbasDB to automatically determine the blob's charset.\n"
+          "Firebird 2.1 *does* provide the required information.\n"
+          "If you call kinterbasdb.init(type_conv=300), KInterbasDB will:\n"
+          " - work perfectly with FB 2.1+\n"
+          " - dangerously assume that the character set of all textual blobs"
+          " is the same as the Connection.charset with FB < 2.1"
+        );
+      goto fail;
+    }
+  }
+
   /* Python will intern most relation and field names, so the following string
    * creation ops should be very efficient. */
   py_relname = PyString_FromStringAndSize(sqlvar->relname, sqlvar->relname_length);
@@ -223,12 +258,17 @@ static PyObject *determine_blob_character_set_id(
     if (con->blob_charset_cache == NULL) { goto fail; }
     might_be_cached = FALSE;
   } else {
-    /* PyDict_GetItem returns a borrowed ref. */
+    /* PyDict_GetItem returns borrowed ref. */
     py_reldict = PyDict_GetItem(con->blob_charset_cache, py_relname);
     might_be_cached = (boolean) (py_reldict != NULL);
     if (might_be_cached) {
       py_char_set_id = PyDict_GetItem(py_reldict, py_fieldname);
-      might_be_cached = (boolean) (py_char_set_id != NULL);
+      if (py_char_set_id != NULL) {
+        might_be_cached = TRUE; /* Was cached. */
+        Py_INCREF(py_char_set_id); /* PyDict_GetItem returned borrowed ref. */
+      } else {
+        might_be_cached = FALSE;
+      }
     }
   }
 
@@ -244,18 +284,20 @@ static PyObject *determine_blob_character_set_id(
         (unsigned char *) PyString_AS_STRING(py_relname);
       unsigned char *fieldname_nullterminated =
         (unsigned char *) PyString_AS_STRING(py_fieldname);
+      /* Transaction_get_handle_p must be called while the GIL is held! */
+      isc_tr_handle *trans_handle_p = Transaction_get_handle_p(trans);
 
       ENTER_GDAL
       isc_blob_lookup_desc(cursor->status_vector,
-          &con->db_handle, CON_GET_TRANS_HANDLE_ADDR(con),
+          &con->db_handle, trans_handle_p,
           relname_nullterminated, fieldname_nullterminated,
           &blob_desc,
           NULL
         );
       LEAVE_GDAL
       if (DB_API_ERROR(cursor->status_vector)) {
-        raise_sql_exception(OperationalError, "Error while attempting to"
-            " determine blob character set: ", cursor->status_vector
+        raise_sql_exception(OperationalError,
+            "determine_blob_character_set_id: ", cursor->status_vector
           );
         goto fail;
       }
@@ -265,7 +307,7 @@ static PyObject *determine_blob_character_set_id(
     {
       ISC_STATUS *sv = con->status_vector;
       isc_db_handle *db_handle = &con->db_handle;
-      isc_tr_handle *trans_handle = CON_GET_TRANS_HANDLE_ADDR(con); /* 2003.10.15a:OK */
+      isc_tr_handle *trans_handle = Transaction_get_handle_p(trans);
 
       const char *charset_determination_statement =
           "SELECT F.RDB$CHARACTER_SET_ID "
@@ -317,7 +359,12 @@ static PyObject *determine_blob_character_set_id(
       ENTER_GDAL
       isc_dsql_allocate_statement(sv, db_handle, &stmt_handle);
       LEAVE_GDAL
-      if (DB_API_ERROR(sv)) { goto fail; }
+      if (DB_API_ERROR(sv)) {
+        raise_sql_exception(OperationalError,
+            "determine_blob_character_set_id: ", sv
+          );
+        goto fail;
+      }
 
       ENTER_GDAL
       isc_dsql_prepare(sv, trans_handle, &stmt_handle,
@@ -325,14 +372,24 @@ static PyObject *determine_blob_character_set_id(
           out_da
         );
       LEAVE_GDAL
-      if (DB_API_ERROR(sv)) { goto fail; }
+      if (DB_API_ERROR(sv)) {
+        raise_sql_exception(OperationalError,
+            "determine_blob_character_set_id: ", sv
+          );
+        goto fail;
+      }
 
       ENTER_GDAL
       isc_dsql_execute2(sv, trans_handle, &stmt_handle,
           SQL_DIALECT_DEFAULT, in_da, out_da
         );
       LEAVE_GDAL
-      if (DB_API_ERROR(sv)) { goto fail; }
+      if (DB_API_ERROR(sv)) {
+        raise_sql_exception(OperationalError,
+            "determine_blob_character_set_id: ", sv
+          );
+        goto fail;
+      }
 
       if (out_var_sqlind == SQLIND_NULL) {
         assert (out_var_charset_id == 0);
@@ -345,45 +402,25 @@ static PyObject *determine_blob_character_set_id(
     }
     #endif
 
-    /* Write the result to the cache before returning it.  Note that from this
-     * point to the end of the cache miss block is unique because we hold
-     * reference ownership over py_char_set_id, and must explicitly release it
-     * if an error occurs. */
+    /* Write the result to the cache before returning it. */
 
     /* This should've been ensured earlier: */
     assert (con->blob_charset_cache != NULL);
 
     assert (py_char_set_id != NULL);
-    #define RELEASE_CHARSETID_AND_FAIL \
-      Py_DECREF(py_char_set_id); \
-      py_char_set_id = NULL; \
-      goto fail;
-
     py_reldict = PyDict_GetItem(con->blob_charset_cache, py_relname);
     if (py_reldict == NULL) {
       /* Create a dictionary for all fields of the relation named py_relname: */
       py_reldict = PyDict_New();
-      if (py_reldict == NULL) {
-        RELEASE_CHARSETID_AND_FAIL;
-      }
+      if (py_reldict == NULL) { goto fail; }
       if (PyDict_SetItem(con->blob_charset_cache, py_relname, py_reldict) != 0) {
         Py_DECREF(py_reldict);
-        RELEASE_CHARSETID_AND_FAIL;
+        goto fail;
       }
-      /* py_reldict is now in the cache, so we can release our reference
-       * without any fear of it being prematurely collected, yet achieve
-       * reference-ownership-symmetry (which is to say, having only a borrowed
-       * reference) with the branch in which py_reldict was already present in
-       * the cache. */
-      Py_DECREF(py_reldict);
     }
     if (PyDict_SetItem(py_reldict, py_fieldname, py_char_set_id) != 0) {
-      RELEASE_CHARSETID_AND_FAIL;
+      goto fail;
     }
-    /* We're supposed to return only a borrowed reference to py_char_set_id, so
-     * here we DECREF it to achieve reference-ownership-symmetry with the
-     * branch in which py_char_set_id was already in the cache. */
-    Py_DECREF(py_char_set_id);
   } /* end cache miss block */
 
   assert (py_char_set_id != NULL);
@@ -391,7 +428,7 @@ static PyObject *determine_blob_character_set_id(
   fail:
     assert (PyErr_Occurred());
     if (py_char_set_id != NULL) {
-      /* Notice that we DO NOT Py_DECREF(py_char_set_id). */
+      Py_DECREF(py_char_set_id);
       py_char_set_id = NULL;
     }
     /* Fall through to clean: */
@@ -400,7 +437,7 @@ static PyObject *determine_blob_character_set_id(
     Py_XDECREF(py_relname);
     Py_XDECREF(py_fieldname);
 
-    return py_char_set_id;
+    return py_char_set_id; /* Returns new reference. */
 } /* determine_blob_character_set_id */
 
 static int get_blob_converter_override_for_direction(
@@ -410,8 +447,9 @@ static int get_blob_converter_override_for_direction(
     boolean *is_unicode_charset
   )
 {
-  /* Both py_converter_override and py_blob_charset_id will hold only
-   * borrowed refs, so no worries about DECREFing them. */
+  /* py_converter_override will hold only a borrowed ref, so client code need
+   * not worries about DECREFing it.  py_blob_charset_id will hold a new ref,
+   * however. */
   int blob_charset_id;
 
   *py_blob_charset_id = determine_blob_character_set_id(cursor, sqlvar);
