@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Streaming Blob Conversion
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -19,7 +19,6 @@
  * without the involvement of a header file. */
 
 static int BlobReader_close_without_unlink(BlobReader *, boolean);
-static int BlobReader_untrack(BlobReader *, boolean);
 
 #ifdef ENABLE_CONNECTION_TIMEOUT
   #define BLOBREADER_REQUIRE_OPEN(self) \
@@ -47,6 +46,29 @@ static int BlobReader_untrack(BlobReader *, boolean);
 
 #define BLOBREADER_BYTES_AVAILABLE(self) ((self)->total_size - (self)->pos)
 
+#define DECLARE_AND_INITIALIZE_TRANS(self) \
+  Transaction *trans; \
+  \
+  assert (self != NULL); \
+  \
+  trans = (self)->trans; \
+  assert (trans != NULL); \
+  assert (Transaction_is_not_closed(trans));
+
+/* Retarded C rule against introducing new declarations except at the beginning
+ * of a scope forces us to repeat code here: */
+#define DECLARE_AND_INITIALIZE_TRANS_AND_SV(self) \
+  Transaction *trans; \
+  ISC_STATUS *sv; \
+  \
+  assert (self != NULL); \
+  \
+  trans = (self)->trans; \
+  assert (trans != NULL); \
+  assert (Transaction_is_not_closed(trans)); \
+  sv = Transaction_get_sv(trans); \
+  assert (sv != NULL);
+
 /*** BlobReaderTracker MEMBER FUNC DEFINITIONS AND SUPPORTING FUNCS: BEGIN ***/
 #include "_kisupport_lifo_linked_list.h"
 
@@ -56,13 +78,48 @@ LIFO_LINKED_LIST_DEFINE_BASIC_METHODS_PYALLOC_NOQUAL(
 /*** BlobReaderTracker MEMBER FUNC DEFINITIONS AND SUPPORTING FUNCS: END *****/
 
 /******************** BlobReader METHODS:BEGIN ********************/
+static void BlobReader_create_references_to_superiors(BlobReader *self,
+    Transaction *trans
+  )
+{
+  assert (self != NULL);
+  assert (self->trans == NULL);
+  assert (self->con_python_wrapper == NULL);
+
+  assert (trans != NULL);
+  Py_INCREF(trans);
+  self->trans = trans;
+
+  /* TAG:TRANSACTION_SUBORDINATE_OBJECT_SURVIVAL_BYPASS: */
+  /* If (Transaction_is_main(self->trans)), the Transaction will not own a
+   * reference to its connection, so we always create an artificial reference,
+   * to ensure that the connection will remain alive at least as long as this
+   * BlobReader does. */
+  self->con_python_wrapper = Transaction_get_con_python_wrapper(trans);
+  assert (self->con_python_wrapper != NULL);
+  Py_INCREF(self->con_python_wrapper);
+} /* BlobReader_set_superior_references */
+
+static void BlobReader_clear_references_to_superiors(BlobReader *self) {
+  assert (self != NULL);
+  assert (self->trans != NULL);
+  assert (self->con_python_wrapper != NULL);
+
+  /* TAG:TRANSACTION_SUBORDINATE_OBJECT_SURVIVAL_BYPASS: */
+  Py_DECREF(self->trans);
+  self->trans = NULL;
+
+  Py_DECREF(self->con_python_wrapper);
+  self->con_python_wrapper = NULL;
+} /* BlobReader_clear_references_to_superiors */
+
 /* This constructor is not accessible to the Python client programmer (i.e.,
  * it's not registered in BlobReaderType's tp_init slot or otherwise exposed)
  * because there's no reason to instantiate BlobReader objects from Python.
  * An attempt to do so raises a TypeError. */
-static BlobReader *BlobReader_create(CConnection *con) {
+static BlobReader *BlobReader_create(Transaction *trans) {
   BlobReader *self = PyObject_New(BlobReader, &BlobReaderType);
-  if (self == NULL) { goto fail_nomem; }
+  if (self == NULL) { goto fail; }
 
   self->state = BLOBREADER_STATE_LIMBO;
 
@@ -74,54 +131,48 @@ static BlobReader *BlobReader_create(CConnection *con) {
   self->pos = -1;
   self->iter_chunk_size = DTT_BLOB_CHUNK_SIZE_UNSPECIFIED;
 
-  /* We deliberately wait to initialize and create new references to anything
-   * related to con until the simpler initializations are known to have
-   * succeeded.
-   * Notice that we only borrow a reference to the CConnection con itself,
-   * but create a new reference to its Python wrapper object, which is an
-   * instance of the Python class kinterbasdb.Connection that's guaranteed to
-   * outlive its companion CConnection. */
-  assert (con != NULL);
-  self->con = con;
-
-  assert (con->python_wrapper_obj != NULL);
-  Py_INCREF(con->python_wrapper_obj);
-  self->con_python_wrapper = con->python_wrapper_obj;
+  self->trans = NULL; /* Initial nullification of mem. */
+  self->con_python_wrapper = NULL;
+  BlobReader_create_references_to_superiors(self, trans);
+  assert (self->trans != NULL);
+  assert (self->con_python_wrapper != NULL);
 
   return self;
-
-  fail_nomem:
+  fail:
     Py_XDECREF(self);
     return NULL;
 } /* BlobReader_create */
 
 static int BlobReader_open(BlobReader *self, ISC_QUAD *blob_id) {
   /* Returns 0 if the op succeeds; -1 on error. */
-  CConnection *con = self->con;
-  assert (con != NULL);
-  assert (self->con_python_wrapper != NULL);
-  assert (self->blob_handle == NULL_BLOB_HANDLE);
+  DECLARE_AND_INITIALIZE_TRANS_AND_SV(self)
+  assert ((self)->blob_handle == NULL_BLOB_HANDLE);
 
   /* Based on the ID of the blob, open a handle to it.  The handle will be used
    * in all future operations on the blob. */
-  ENTER_GDAL
-  isc_open_blob2(con->status_vector,
-      &con->db_handle, &con->trans_handle,
-      &self->blob_handle, blob_id,
-      /* Last two params indicate "no blob parameter buffer supplied": */
-      0, NULL
-    );
-  LEAVE_GDAL
-  if (DB_API_ERROR(con->status_vector)) {
+  {
+    /* Notice that we call Transaction_get_handle_p while holding the GIL. */
+    isc_db_handle *db_handle_p = Transaction_get_db_handle_p(trans);
+    isc_tr_handle *tr_handle_p = Transaction_get_handle_p(trans);
+
+    ENTER_GDAL
+    isc_open_blob2(sv, db_handle_p, tr_handle_p,
+        &self->blob_handle, blob_id,
+        /* Last two params indicate "no blob parameter buffer supplied": */
+        0, NULL
+      );
+    LEAVE_GDAL
+  }
+  if (DB_API_ERROR(sv)) {
     raise_sql_exception(OperationalError, "BlobReader_open.isc_open_blob2: ",
-        con->status_vector
+        sv
       );
     return -1;
   }
 
   /* Determine the total size and max segment size of the blob. */
   if (0 !=
-      _blob_info_total_size_and_max_segment_size(con->status_vector,
+      _blob_info_total_size_and_max_segment_size(sv,
           &self->blob_handle, &self->total_size, &self->max_segment_size
         )
     )
@@ -129,7 +180,7 @@ static int BlobReader_open(BlobReader *self, ISC_QUAD *blob_id) {
   assert (self->total_size >= 0);
   /* max_segment_size is guaranteed to be >= 0 because it's unsigned. */
 
-  if (BlobReaderTracker_add(&self->con->open_blobreaders, self) == -1) {
+  if (BlobReaderTracker_add(&trans->open_blobreaders, self) == -1) {
     return -1;
   }
 
@@ -143,28 +194,51 @@ static int _BlobReader_close_handle_only(
   )
 {
   /* Returns 0 if the op succeeds; -1 on error. */
-  assert (self->con != NULL);
-  assert (self->con_python_wrapper != NULL);
+  int status = -1;
+
+  DECLARE_AND_INITIALIZE_TRANS_AND_SV(self)
   assert (self->blob_handle != NULL_BLOB_HANDLE);
 
-  ENTER_GDAL
-  isc_close_blob(self->con->status_vector, &self->blob_handle);
-  LEAVE_GDAL
-  if (!DB_API_ERROR(self->con->status_vector)) {
+  {
+    /* This code can be reached when the CTT is timing out a connection.  In
+     * that case, we want the GIL to remain held during the entire timeout
+     * operation. */
+    OPEN_LOCAL_GIL_MANIPULATION_INFRASTRUCTURE
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    const boolean should_manip_gil = NOT_RUNNING_IN_CONNECTION_TIMEOUT_THREAD;
+    if (should_manip_gil) {
+    #endif
+      LEAVE_GIL_WITHOUT_AFFECTING_DB_AND_WITHOUT_STARTING_CODE_BLOCK
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    }
+    #endif
+    ENTER_GDAL_WITHOUT_LEAVING_PYTHON
+
+    isc_close_blob(sv, &self->blob_handle);
+
+    LEAVE_GDAL_WITHOUT_ENTERING_PYTHON
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    if (should_manip_gil) {
+    #endif
+      ENTER_GIL_WITHOUT_AFFECTING_DB_AND_WITHOUT_ENDING_CODE_BLOCK
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    }
+    #endif
+    CLOSE_LOCAL_GIL_MANIPULATION_INFRASTRUCTURE
+  } /* end of lock manipulation scope */
+
+  if (!DB_API_ERROR(sv)) {
     assert (self->blob_handle == NULL_BLOB_HANDLE);
+    status = 0;
   } else {
-    raise_sql_exception(OperationalError, "_BlobReader_close.isc_close_blob: ",
-        self->con->status_vector
-      );
     if (allowed_to_raise) {
-      return -1;
+      raise_sql_exception(OperationalError, "_BlobReader_close: ", sv);
     } else {
-      SUPPRESS_EXCEPTION;
       self->blob_handle = NULL_BLOB_HANDLE;
     }
   }
 
-  return 0;
+  return status;
 } /* _BlobReader_close_handle_only */
 
 static int _BlobReader_close(BlobReader *self,
@@ -174,12 +248,13 @@ static int _BlobReader_close(BlobReader *self,
   /* Returns 0 if the op succeeds; -1 on error. */
   int status = 0;
 
+  DECLARE_AND_INITIALIZE_TRANS(self)
+
   assert (BlobReader_is_open(self));
-  assert (self->con != NULL);
   #ifdef ENABLE_CONNECTION_TIMEOUT
     assert (
-          Connection_timeout_enabled(self->con)
-        ? CURRENT_THREAD_OWNS_TP(self->con->timeout)
+          Connection_timeout_enabled(Transaction_get_con(trans))
+        ? CURRENT_THREAD_OWNS_CON_TP(Transaction_get_con(trans))
         : TRUE
       );
   #endif /* ENABLE_CONNECTION_TIMEOUT */
@@ -191,17 +266,14 @@ static int _BlobReader_close(BlobReader *self,
     if (allowed_to_raise) {
       goto fail;
     } else {
-      SUPPRESS_EXCEPTION;
       self->blob_handle = NULL_BLOB_HANDLE;
+      SUPPRESS_EXCEPTION;
     }
   }
   assert (self->blob_handle == NULL_BLOB_HANDLE);
 
   if (should_unlink_self) {
-    if (BlobReaderTracker_remove(&self->con->open_blobreaders, self, TRUE)
-        != 0
-       )
-    {
+    if (BlobReaderTracker_remove(&trans->open_blobreaders, self, TRUE) != 0) {
       status = -1;
       if (allowed_to_raise) {
         goto fail;
@@ -214,20 +286,12 @@ static int _BlobReader_close(BlobReader *self,
   self->pos = -1;
   self->state = BLOBREADER_STATE_CLOSED;
 
-  /* Since this BlobReader's blob handle is now closed, it'll never again
-   * need access to the Connection's database handle or transaction handle.
-   * We can therefore discard our Connection immediately.  Note that we *own*
-   * the reference to the Python wrapper around the CConnection, but not the
-   * one to the CConnection itself. */
-  Py_DECREF(self->con_python_wrapper);
-  self->con_python_wrapper = NULL;
-  self->con = NULL;
-
   goto clean;
   fail:
     assert (allowed_to_raise ? !!PyErr_Occurred() : !PyErr_Occurred());
     /* Fall through to clean: */
   clean:
+    assert (!allowed_to_raise ? self->state == BLOBREADER_STATE_CLOSED : TRUE);
     return status;
 } /* _BlobReader_close */
 
@@ -239,8 +303,21 @@ static int BlobReader_close_without_unlink(
 } /* BlobReader_close_without_unlink */
 
 static int BlobReader_untrack(BlobReader *self, boolean allowed_to_raise) {
-  /* Same as BlobReader_close_without_unlink. */
-  return BlobReader_close_without_unlink(self, allowed_to_raise);
+  int res = -1;
+  assert (self->trans != NULL);
+  assert (self->con_python_wrapper != NULL);
+
+  assert (BlobReader_is_open(self));
+
+  if (BlobReader_close_without_unlink(self, allowed_to_raise) == 0) {
+    res = 0;
+  }
+
+  /* 2007.01.17: Closing a BlobReader no longer clears its superior refs: */
+  assert (self->trans != NULL);
+  assert (self->con_python_wrapper != NULL);
+  assert (!allowed_to_raise ? self->state != BLOBREADER_STATE_OPEN : TRUE);
+  return res;
 } /* BlobReader_close_untrack */
 
 static int BlobReader_close_with_unlink(
@@ -253,37 +330,35 @@ static int BlobReader_close_with_unlink(
   return _BlobReader_close(self, TRUE, allowed_to_raise);
 } /* BlobReader_close_with_unlink */
 
-
 static PyObject *pyob_BlobReader_close(BlobReader *self) {
   /* Thin Python wrapper over BlobReader_close_with_unlink. */
   PyObject *ret = NULL;
+  CConnection *con = NULL;
 
-  PyObject *con_wrapper;
-  CConnection *con;
   #ifdef ENABLE_CONNECTION_TIMEOUT
     boolean con_timeout_was_enabled;
     boolean con_activation_succeeded = TRUE;
   #endif /* ENABLE_CONNECTION_TIMEOUT */
 
   BLOBREADER_REQUIRE_OPEN(self);
-  con = self->con;
-  assert (con != NULL);
-  assert (con->python_wrapper_obj != NULL);
+
+  assert (self->trans != NULL);
   assert (self->con_python_wrapper != NULL);
-  assert (self->con_python_wrapper == con->python_wrapper_obj);
-  /* We need to make sure the connection stays alive long enough to be unlocked
-   * at the end of this function: */
-  con_wrapper = self->con_python_wrapper;
-  Py_INCREF(con_wrapper);
+  con = Transaction_get_con(self->trans);
+  assert (con != NULL);
 
   #ifdef ENABLE_CONNECTION_TIMEOUT
-    con_timeout_was_enabled = (boolean) Connection_timeout_enabled(con);
+    {
+      con_timeout_was_enabled = (boolean) Connection_timeout_enabled(con);
 
-    if (con_timeout_was_enabled) {
-      ACQUIRE_CON_TP_WITH_GIL_HELD(con);
-      CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION__ALREADY_LOCKED(con,
-          goto fail_without_passivating
-        );
+      if (con_timeout_was_enabled) {
+        assert (!CURRENT_THREAD_OWNS_CON_TP(con));
+        ACQUIRE_CON_TP_WITH_GIL_HELD(con);
+        CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION__ALREADY_LOCKED(
+            con, goto fail_without_passivating
+          );
+        assert (CURRENT_THREAD_OWNS_CON_TP(con));
+      }
     }
   #endif /* ENABLE_CONNECTION_TIMEOUT */
 
@@ -294,7 +369,9 @@ static PyObject *pyob_BlobReader_close(BlobReader *self) {
   } else {
     goto fail;
   }
-  assert (self->con == NULL);
+  /* 2007.01.17: Closing a BlobReader no longer clears its superior refs: */
+  assert (self->trans != NULL);
+  assert (self->con_python_wrapper != NULL);
 
   goto clean;
   #ifdef ENABLE_CONNECTION_TIMEOUT
@@ -304,19 +381,22 @@ static PyObject *pyob_BlobReader_close(BlobReader *self) {
   #endif /* ENABLE_CONNECTION_TIMEOUT */
   fail:
     assert (PyErr_Occurred());
+    assert (ret == NULL);
+    assert (con_activation_succeeded);
     /* Fall through to clean: */
   clean:
     #ifdef ENABLE_CONNECTION_TIMEOUT
       if (con_timeout_was_enabled) {
         if (con_activation_succeeded) {
+          assert (CURRENT_THREAD_OWNS_CON_TP(con));
           CON_PASSIVATE__ALREADY_LOCKED(con);
         }
         TP_UNLOCK(con->timeout);
+        assert (!CURRENT_THREAD_OWNS_CON_TP(con));
       }
     #endif /* ENABLE_CONNECTION_TIMEOUT */
-    Py_DECREF(con_wrapper);
-
     CON_MUST_NOT_BE_ACTIVE(con);
+
     return ret;
 } /* pyob_BlobReader_close */
 
@@ -325,8 +405,8 @@ static PyObject *BlobReader_read(BlobReader *self, int req_chunk_size) {
 
   /* The caller should've already ensured that self is open: */
   assert (BlobReader_is_open(self));
-  assert (self->con != NULL);
-  CON_MUST_ALREADY_BE_ACTIVE(self->con);
+  assert (self->trans != NULL);
+  CON_MUST_ALREADY_BE_ACTIVE(Transaction_get_con(self->trans));
 
   /* If the requested number of bytes was negative (including the default value
    * of -1), read all bytes available.
@@ -352,8 +432,8 @@ static PyObject *BlobReader_read(BlobReader *self, int req_chunk_size) {
      * subsequent call to isc_get_segment will start where the previous one
      * left off, which is exactly what we need. */
     PyObject *py_str = conv_out_blob_materialized_in_single_chunk(
-        self->con->status_vector, &self->blob_handle, self->max_segment_size,
-        req_chunk_size, TRUE
+        Transaction_get_sv(self->trans), &self->blob_handle,
+        self->max_segment_size, req_chunk_size, TRUE
       );
     if (py_str == NULL) { return NULL; }
     self->pos += req_chunk_size;
@@ -367,9 +447,11 @@ static PyObject *pyob_BlobReader_read(BlobReader *self, PyObject *args) {
   int req_chunk_size = -1;
 
   BLOBREADER_REQUIRE_OPEN(self);
-  assert (self->con != NULL);
+  assert (self->trans != NULL);
 
-  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(self->con, return NULL);
+  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(
+      Transaction_get_con(self->trans), return NULL
+    );
 
   if (!PyArg_ParseTuple(args, "|i", &req_chunk_size)) { goto fail; }
 
@@ -382,8 +464,11 @@ static PyObject *pyob_BlobReader_read(BlobReader *self, PyObject *args) {
     assert (ret == NULL);
     /* Fall through to clean: */
   clean:
-    CON_PASSIVATE(self->con);
-    CON_MUST_NOT_BE_ACTIVE(self->con);
+    {
+      CConnection *con = Transaction_get_con(self->trans);
+      CON_PASSIVATE(con);
+      CON_MUST_NOT_BE_ACTIVE(con);
+    }
     return ret;
 } /* pyob_BlobReader_read */
 
@@ -393,9 +478,11 @@ static PyObject *pyob_BlobReader__iter_read_chunk(BlobReader *self) {
   PyObject *ret = NULL;
 
   BLOBREADER_REQUIRE_OPEN(self);
-  assert (self->con != NULL);
+  assert (self->trans != NULL);
 
-  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(self->con, return NULL);
+  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(
+      Transaction_get_con(self->trans), return NULL
+    );
 
   if (self->iter_chunk_size == DTT_BLOB_CHUNK_SIZE_UNSPECIFIED) {
     raise_exception(ProgrammingError, "This method is private and must not be"
@@ -413,8 +500,11 @@ static PyObject *pyob_BlobReader__iter_read_chunk(BlobReader *self) {
     assert (ret == NULL);
     /* Fall through to clean: */
   clean:
-    CON_PASSIVATE(self->con);
-    CON_MUST_NOT_BE_ACTIVE(self->con);
+    {
+      CConnection *con = Transaction_get_con(self->trans);
+      CON_PASSIVATE(con);
+      CON_MUST_NOT_BE_ACTIVE(con);
+    }
     return ret;
 } /* pyob_BlobReader__iter_read_chunk */
 
@@ -426,9 +516,11 @@ static PyObject *pyob_BlobReader_chunks(BlobReader *self, PyObject *args) {
   int req_chunk_size = -1;
 
   BLOBREADER_REQUIRE_OPEN(self);
-  assert (self->con != NULL);
+  assert (self->trans != NULL);
 
-  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(self->con, return NULL);
+  CON_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(
+      Transaction_get_con(self->trans), return NULL
+    );
 
   if (self->iter_chunk_size != DTT_BLOB_CHUNK_SIZE_UNSPECIFIED) {
     raise_exception(ProgrammingError, "At most one iterator can be opened on"
@@ -463,7 +555,6 @@ static PyObject *pyob_BlobReader_chunks(BlobReader *self, PyObject *args) {
 
     Py_DECREF(bound_method__iter_read_chunk);
     Py_DECREF(sentinel);
-
   }
 
   goto clean;
@@ -472,8 +563,11 @@ static PyObject *pyob_BlobReader_chunks(BlobReader *self, PyObject *args) {
     assert (it == NULL);
     /* Fall through to clean: */
   clean:
-    CON_PASSIVATE(self->con);
-    CON_MUST_NOT_BE_ACTIVE(self->con);
+    {
+      CConnection *con = Transaction_get_con(self->trans);
+      CON_PASSIVATE(con);
+      CON_MUST_NOT_BE_ACTIVE(con);
+    }
     return it;
 } /* pyob_BlobReader_chunks */
 
@@ -513,53 +607,95 @@ static PyObject *pyob_BlobReader_repr(BlobReader *self) {
 } /* pyob_BlobReader_repr */
 
 static void pyob_BlobReader___del__(BlobReader *self) {
-  #define BR_CLEAR_CON_REFS(br) \
-    if ((br)->con != NULL) { \
-      (br)->con = NULL; /* We don't own the ref. */ \
-    } \
-    if ((br)->con_python_wrapper != NULL) { \
-      Py_DECREF((br)->con_python_wrapper); \
-      (br)->con_python_wrapper = NULL; \
-    }
-
-  CConnection *con = self->con;
   assert (NOT_RUNNING_IN_CONNECTION_TIMEOUT_THREAD);
 
-  if (con != NULL) {
-    PyObject *con_python_wrapper = con->python_wrapper_obj;
-    /* If the ConnectionTimeoutThread tries to indirectly close self at the
-     * same time this function is trying to, we need to synchronize. */
-    Py_XINCREF(con_python_wrapper);
-    Py_INCREF(con);
+  if (self->trans != NULL) {
+    /* TAG:TRANSACTION_SUBORDINATE_OBJECT_SURVIVAL_BYPASS: */
+    /* We need to make sure that the Transaction and the connection that
+     * underlies it stay alive at least until we've completely the process of
+     * removing their references to self.
+     *
+     * Fortunately, as of the changes on 2007.01.17, this is easy:
+     * BlobReader_create_references_to_superiors creates references to *both*
+     * the Transaction and the kinterbasdb.Connection that underlies it.
+     * BlobReader_create_references_to_superiors creates that artificial
+     * reference to the kinterbasdb.Connection in case
+     *   Transaction_is_main(self->trans)
+     * , which would mean that self->trans doesn't actually hold a physical
+     * reference to its connection.
+     *
+     * BlobReader_clear_references_to_superiors, which is not invoked until the
+     * end of this very code block, clears both of those references in the
+     * appropriate order.
+     *
+     * So, we do not need to do anything special to ensure that the Transaction
+     * survives long enough for self to be torn down in an orderly manner. */
+    CConnection *con;
+    Transaction *trans = self->trans;
 
-    #ifdef ENABLE_CONNECTION_TIMEOUT
-      ACQUIRE_CON_TP_WITH_GIL_HELD(con);
-    #endif /* ENABLE_CONNECTION_TIMEOUT */
+    assert (trans->ob_refcnt >= 1);
 
-    if (BlobReader_is_open(self)) {
-      if (BlobReader_close_with_unlink(self, FALSE) == 0) {
-        assert (self->con == NULL);
-        assert (self->con_python_wrapper == NULL);
-      } else {
-        BR_CLEAR_CON_REFS(self);
+    assert (self->con_python_wrapper != NULL);
+    /* We can make the following assertion about the kinterbasdb.Connection
+     * instance's reference count because we know that self owns a reference
+     * *directly* to the kinterbasdb.Connectioin, regardless of whether
+     * self->trans actually owns a physical reference to it (see comment above
+     * for details): */
+    assert (self->con_python_wrapper->ob_refcnt >= 1);
+
+    con = Transaction_get_con(trans);
+    assert (con == NULL ? !BlobReader_is_open(self) : TRUE);
+    if (con != NULL) {
+      #ifdef ENABLE_CONNECTION_TIMEOUT
+        assert (
+              Connection_timeout_enabled(con)
+            ? !CURRENT_THREAD_OWNS_CON_TP(con)
+            : TRUE
+          );
+        ACQUIRE_CON_TP_WITH_GIL_HELD(con);
+        assert (
+              Connection_timeout_enabled(con)
+            ? CURRENT_THREAD_OWNS_CON_TP(con)
+            : TRUE
+          );
+      #endif /* ENABLE_CONNECTION_TIMEOUT */
+
+      if (BlobReader_is_open(self)) {
+        BlobReader_close_with_unlink(self, FALSE);
       }
-    } else {
-      BR_CLEAR_CON_REFS(self);
-    }
+      /* As of 2007.01.17, BlobReader_close_with_unlink should no longer clear
+       * self->trans: */
+      assert (self->trans != NULL);
+      assert (self->trans == trans);
+      assert (trans->ob_refcnt >= 1);
+      assert (self->con_python_wrapper != NULL);
 
-    #ifdef ENABLE_CONNECTION_TIMEOUT
-      RELEASE_CON_TP(con);
-    #endif /* ENABLE_CONNECTION_TIMEOUT */
+      #ifdef ENABLE_CONNECTION_TIMEOUT
+        assert (
+              Connection_timeout_enabled(con)
+            ? CURRENT_THREAD_OWNS_CON_TP(con)
+            : TRUE
+          );
+        RELEASE_CON_TP(con);
+        assert (
+              Connection_timeout_enabled(con)
+            ? !CURRENT_THREAD_OWNS_CON_TP(con)
+            : TRUE
+          );
+      #endif /* ENABLE_CONNECTION_TIMEOUT */
+    } /* end of if (con != NULL) block */
 
-    Py_DECREF(con);
-    Py_XDECREF(con_python_wrapper);
-    con = NULL;
-  }
+    BlobReader_clear_references_to_superiors(self);
+    assert (self->trans == NULL);
+    assert (self->con_python_wrapper == NULL);
+    /* Note that if more code is ever added after this point, IT IS NO LONGER
+     * SAFE TO REFER TO trans OR con, so code such as this should be enabled:
+     *   trans = NULL;
+     *   con = NULL;   */
+  } /* end of if (self->trans != NULL) block */
 
   assert (!BlobReader_is_open(self));
-  /* Closing the BlobReader should have discarded and nullified our Connection
-   * refs: */
-  assert (self->con == NULL);
+  assert (self->trans == NULL);
   assert (self->con_python_wrapper == NULL);
   assert (self->blob_handle == NULL_BLOB_HANDLE);
 
