@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Parameter Conversion DB->Py
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -208,7 +208,5 @@ static PyObject *conv_out_time(char *data) {
 
 #endif /* INTERBASE_6_OR_LATER */
 
-#ifdef INTERBASE_7_OR_LATER
 #define conv_out_boolean(data) \
   PyBool_FromLong( *((ISC_BOOLEAN *) (data)) )
-#endif /* INTERBASE_7_OR_LATER */
