@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Parameter Conversion Py->DB
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -635,8 +635,8 @@ static InputStatus conv_in_blob_materialized(
    * present, Python strings/buffers and database blobs have the same maximum
    * size:  2147483647 bytes. */
   ISC_STATUS *status_vector = cursor->status_vector;
-  isc_db_handle db_handle = cursor->connection->db_handle;
-  isc_tr_handle trans_handle = CON_GET_TRANS_HANDLE(cursor->connection); /* 2003.10.15a:OK */
+  isc_db_handle db_handle = *Transaction_get_db_handle_p(cursor->trans);
+  isc_tr_handle trans_handle = *Transaction_get_handle_p(cursor->trans);
 
   /* Next statement allocates space for the blob's id, not for the blob's
    * contents (the contents are written segment-at-a-time in
@@ -678,8 +678,6 @@ static InputStatus conv_in_blob_materialized(
 } /* conv_in_blob_materialized */
 
 
-#ifdef INTERBASE_7_OR_LATER
-
 #define conv_in_boolean_conventional(py_input, sqlvar) \
     _conv_in_boolean(FALSE, py_input, &(sqlvar)->sqldata)
 
@@ -700,8 +698,6 @@ static InputStatus _conv_in_boolean(boolean is_array_element,
     return INPUT_ERROR;
 } /* conv_in_boolean */
 
-#endif /* INTERBASE_7_OR_LATER */
-
 
 /******************** UTILITY FUNCTIONS:BEGIN ********************/
 
