@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Dynamic Type Translation
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -38,6 +38,8 @@ PyObject *cached_type_name_TIMESTAMP;
 PyObject *cached_type_name_DATE;
 PyObject *cached_type_name_TIME;
 
+PyObject *cached_type_name_BOOLEAN;
+
 #define DTT_DISALLOW_POSITIONAL -1
 #define DTT_POSITIONAL_IN_RANGE(i) (i >= 0 && i <= SHRT_MAX)
 
@@ -74,6 +76,8 @@ static int init_kidb_type_translation(void) {
   _ICTN( cached_type_name_DATE,           "DATE"                            );
   _ICTN( cached_type_name_TIME,           "TIME"                            );
 
+  _ICTN( cached_type_name_BOOLEAN,        "BOOLEAN"                         );
+
   return 0;
 
   fail:
@@ -130,6 +134,9 @@ static PyObject *_get_cached_type_name_for_conventional_code(
       return cached_type_name_TIME;
     #endif /* INTERBASE_6_OR_LATER */
 
+    case SQL_BOOLEAN:
+      return cached_type_name_BOOLEAN;
+
     default:
       {
         PyObject *err_msg = PyString_FromFormat(
@@ -191,6 +198,9 @@ static PyObject *_get_cached_type_name_for_array_code(
       return cached_type_name_TIME;
     #endif /* INTERBASE_6_OR_LATER */
 
+    case blr_boolean_dtype:
+      return cached_type_name_BOOLEAN;
+
     case blr_quad:
       /* The database engine does not support arrays of arrays. */
     case blr_blob:
@@ -309,7 +319,7 @@ _make__connection_get_DIRECTION_converter(in)
   { \
     PyObject *trans_dict = cursor->type_trans_ ## direction; \
     PyObject *converter = _get_converter(trans_dict, sqlvar_index, \
-        cursor->connection->dialect, \
+        Transaction_get_dialect(cursor->trans), \
         data_type, data_subtype, scale, is_array_field \
       ); \
     \
@@ -318,7 +328,8 @@ _make__connection_get_DIRECTION_converter(in)
     } \
     \
     /* Fall back on the connection's translation dictionary, if any. */ \
-    return connection_get_ ## direction ## _converter(cursor->connection, \
+    return connection_get_ ## direction ## _converter( \
+        Transaction_get_con(cursor->trans), \
         data_type, data_subtype, scale, is_array_field \
       ); \
   }
@@ -342,7 +353,7 @@ _make__cursor_get_DIRECTION_converter(out)
       } \
     } \
     /* Fall back on the connection's translation dictionary, if any. */ \
-    trans_dict = cursor->connection->type_trans_ ## direction; \
+    trans_dict = Transaction_get_con(cursor->trans)->type_trans_ ## direction; \
     if (trans_dict != NULL) { \
       converter = PyDict_GetItem(trans_dict, type_name); \
       if (converter != NULL || PyErr_Occurred()) { \
@@ -403,8 +414,8 @@ static PyObject *cursor_get_translator_output_type(
       } /* Else, fall through and search the connection's output type dict. */
     }
 
-    return connection_get_translator_output_type(cursor->connection,
-        translator_key
+    return connection_get_translator_output_type(
+        Transaction_get_con(cursor->trans), translator_key
       );
   }
 } /* cursor_get_translator_output_type */
