@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Connection Timeout
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -18,6 +18,9 @@
 #include "_kicore_connection_timeout.h"
 
 /*************************** PRELIMINARIES: BEGIN ****************************/
+static double ConnectionTimeoutParams_active_secs(ConnectionTimeoutParams *tp);
+static double ConnectionTimeoutParams_idle_secs(ConnectionTimeoutParams *tp);
+
 /* Inform the ConnectionTimeoutThread that a new connection has been
  * added to the tracker: */
 #ifdef PLATFORM_WINDOWS
@@ -404,10 +407,6 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
     assert (con != NULL);
     assert (con->ob_refcnt > 0);
     assert (con->state == CON_STATE_OPEN);
-    /* Connections that have timeout enabled are not supposed to be allowed to
-     * participate in a distributed transaction group.  The Python class
-     * ConnectionGroup should have enforced this: */
-    assert (con->group == NULL);
 
     tp = con->timeout;
     assert (tp != NULL);
@@ -422,10 +421,8 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
         boolean should_time_out = TRUE;
         boolean timeout_was_transparent = FALSE;
         boolean had_transaction;
-        double active_secs =
-          ((double) (tp->last_active - tp->connected_at)) / 1000.0;
-        double idle_secs =
-          ((double) (time_millis() - tp->last_active)) / 1000.0;
+        const double active_secs = ConnectionTimeoutParams_active_secs(tp);
+        const double idle_secs = ConnectionTimeoutParams_idle_secs(tp);
 
         PyObject *py_dsn = NULL;
         PyObject *py_active_secs = NULL;
@@ -444,15 +441,14 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
          * relatively rarely. */
         ENTER_GIL_USING_THREADSTATE(tstate);
 
-        /* GIL must be held before CON_HAS_TRANSACTION is called: */
-        had_transaction = CON_HAS_TRANSACTION(DV_CCON(con));
+        /* GIL must be held before Connection_has_any_open_transaction is
+         * called: */
+        had_transaction = Connection_has_any_open_transaction(DV_CCON(con));
 
         assert (con->dsn != NULL && con->dsn_len > 0);
         if (tp->py_callback_before != NULL || tp->py_callback_after != NULL) {
           py_dsn = PyString_FromStringAndSize(con->dsn, con->dsn_len);
-          if (py_dsn == NULL) {
-            SUPPRESS_EXCEPTION;
-          }
+          if (py_dsn == NULL) { SUPPRESS_EXCEPTION; }
 
           /* py_active_secs and py_idle_secs will be checked for successful
            * construction later. */
@@ -549,51 +545,34 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
                       if (!had_transaction) {
                         /* If the connection didn't have a transaction
                          * originally, it shouldn't now. */
-                        assert (!CON_HAS_TRANSACTION(DV_CCON(con)));
-                        timeout_was_transparent = (verdict != CT_NONTRANSPARENT);
+                        assert (!Connection_has_any_open_transaction(
+                            DV_CCON(con)
+                          ));
+                        timeout_was_transparent =
+                          (verdict != CT_NONTRANSPARENT);
                       } else {
-                        boolean resolution_successful;
-                        WhichTransactionOperation op = (
+                        const WhichTransactionOperation op = (
                             verdict == CT_COMMIT ? OP_COMMIT : OP_ROLLBACK
                           );
-                        /* For an explanation of what we're doing here with the
-                         * BlobReaders, see the comments in the
-                         * Connection_copy_BlobReader_pointers function. */
-                        BlobReader **blob_readers = NULL;
-                        Py_ssize_t n_blobreaders = -1;
+                        const TransactionalOperationResult trans_res_status =
+                          Connection_resolve_all_transactions_from_CTT(con, op);
+                        /* Connection_resolve_all_transactions_from_CTT should
+                         * have already suppressed any Python exception: */
+                        assert (!PyErr_Occurred());
 
-                        if (con->open_blobreaders != NULL) {
-                          blob_readers = Connection_copy_BlobReader_pointers(
-                              con, &n_blobreaders
-                            );
-                          if (blob_readers != NULL) {
-                            assert (n_blobreaders != -1);
-                          } else {
-                            /* Even if the attempt to copy the BlobReader
-                             * pointers fails, we don't want to give up on our
-                             * attempt to resolve the transaction. */
-                            SUPPRESS_EXCEPTION;
-                          }
-                        }
-
-                        resolution_successful = (boolean) (OP_RESULT_OK ==
-                            commit_or_rollback(op, DV_CCON(con), FALSE)
-                          );
-                        if (resolution_successful) {
-                          assert (!CON_HAS_TRANSACTION(DV_CCON(con)));
-                          timeout_was_transparent = (verdict != CT_NONTRANSPARENT);
-
-                          if (blob_readers != NULL) {
-                            Connection_former_BlobReaders_flag_timeout_and_free
-                              (blob_readers, n_blobreaders);
-                            blob_readers = NULL;
-                          }
+                        if (trans_res_status == OP_RESULT_OK) {
+                          assert (!Connection_has_any_open_transaction(
+                              DV_CCON(con)
+                            ));
+                          timeout_was_transparent =
+                            (verdict != CT_NONTRANSPARENT);
                         } else {
-                          SUPPRESS_EXCEPTION;
+                          assert (!timeout_was_transparent);
                         }
                       }
                       assert (should_time_out);
                       break;
+
                     default:
                       /* This should never be reached, because verdict should
                        * not have received the value of verdict_L unless that
@@ -601,7 +580,7 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
                       assert (FALSE);
                   }
                   Py_DECREF(py_callback_res);
-                }
+                } /* end of is-py_callback_res-null block */
               } else {
                 SUPPRESS_EXCEPTION;
                 continue_callback_attempt = FALSE;
@@ -610,17 +589,17 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
               /* If an exception arose, it should've been cleared so the sweep
                * can continue. */
               assert (!PyErr_Occurred());
-            }
-          }
+            } /* end of is-py_callback_dict-null block */
+          } /* end of initial should-continue_callback_attempt block */
 
           Py_XDECREF(py_callback_dict);
           Py_XDECREF(py_has_transaction);
-        }
+        } /* end of did-user-supply-callback block */
 
         /* If the callback did not veto the timeout, close the connection.  Use
          * Connection_close_from_CTT so that this thread doesn't need to
          * release global_ctm.lock during the connection's closure.  This
-         * thread remove the connection from the tracker manually (see block
+         * thread removes the connection from the tracker manually (see block
          * below, after GIL release). */
         if (should_time_out) {
           ++n_cons_tried_to_time_out;
@@ -692,7 +671,7 @@ static int CTM_apply_timeout(PyThreadState *tstate) {
                        con_timeout__s_idle_secs, py_idle_secs            ) == 0
                  )
               {
-                assert (!CON_HAS_TRANSACTION(DV_CCON(con)));
+                assert (!Connection_has_any_open_transaction(DV_CCON(con)));
                 {
                   PyObject *py_callback_res = PyObject_CallFunctionObjArgs(
                         DV_PYO(tp->py_callback_after), py_callback_dict, NULL
@@ -1178,17 +1157,27 @@ static void _ConnectionTimeoutParams_touch(ConnectionTimeoutParams *tp) {
   tp->last_active = time_millis();
   tp->soonest_might_time_out = tp->last_active + tp->timeout_period;
 } /* _ConnectionTimeoutParams_touch */
+
+static double ConnectionTimeoutParams_active_secs(ConnectionTimeoutParams *tp) {
+  return ((double) (tp->last_active - tp->connected_at)) / 1000.0;
+} /* ConnectionTimeoutParams_active_secs */
+
+static double ConnectionTimeoutParams_idle_secs(ConnectionTimeoutParams *tp) {
+  return ((double) (time_millis() - tp->last_active)) / 1000.0;
+} /* ConnectionTimeoutParams_idle_secs */
+
 /**** ConnectionTimeoutParams MEMBER FUNC DEFS AND SUPPORTING FUNCS: END *****/
 
 /****** CConnection ACTIVATION AND DEACTIVATION INFRASTRUCTURE: BEGIN ********/
 static int Connection_activate(CConnection *con,
-    boolean con_tp_already_locked,
-    boolean allow_transparent_resumption
+    const boolean con_tp_already_locked,
+    const boolean allow_transparent_resumption
   )
 {
   /* The GIL must be held when this function is called. */
   int status = 0;
 
+
   if (!Connection_timeout_enabled(con)) {
     if (con->state != CON_STATE_OPEN) {
       raise_exception(ProgrammingError, "Invalid connection state.  The"
@@ -1199,13 +1188,22 @@ static int Connection_activate(CConnection *con,
   } else {
     ConnectionTimeoutParams *tp = con->timeout;
     ConnectionOpState achieved_state;
+
+    assert (tp != NULL);
     assert (NOT_RUNNING_IN_CONNECTION_TIMEOUT_THREAD);
 
+    /* In checked build, verify that the caller's claim about lock ownership
+     * (delivered via boolean con_tp_already_locked) matches reality: */
+    assert (
+          con_tp_already_locked
+        ?  CURRENT_THREAD_OWNS_CON_TP(con)
+        : !CURRENT_THREAD_OWNS_CON_TP(con)
+      );
     if (!con_tp_already_locked) {
       ACQUIRE_TP_WITH_GIL_HELD(tp);
     }
-
     assert (CURRENT_THREAD_OWNS_CON_TP(con));
+
     achieved_state = ConnectionTimeoutParams_trans_while_already_locked(tp,
         CONOP_IDLE, CONOP_ACTIVE
       );
