@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Cursor
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -24,7 +24,6 @@ static int PSTracker_remove(PSTracker **, PreparedStatement *cont, boolean);
 
 static int PreparedStatement_isc_close(PreparedStatement *, boolean);
 
-static boolean _check_statement_length(Py_ssize_t);
 static PyObject *_generic_single_item_isc_dsql_sql_info_request(
     isc_stmt_handle *, ISC_STATUS *, const char, const short
   );
@@ -37,8 +36,10 @@ typedef int (*PSCacheMappedFunction)(
 #define PSCache_has_been_deleted(self) \
   ((self)->container == NULL)
 
+/* Note that PS_STATE_CLOSED has a very different meaning from the "closed"
+ * states of most other objects: */
 #define PreparedStatement_is_open(self) \
-  ((self)->state != PS_STATE_DROPPED && (self)->state != PS_STATE_CREATED)
+  ((self)->state == PS_STATE_OPEN || (self)->state == PS_STATE_CLOSED)
 
 /* Python strings are immutable, and Python's garbage collector does not
  * relocate memory, so if two PyObject pointers point to the same memory
@@ -99,7 +100,13 @@ static int PSCache_append(PSCache *self, PreparedStatement *ps) {
 
   assert (ps != NULL);
   assert (PreparedStatement_is_open(ps));
+
+  /* Any PreparedStatement added to a Cursor's PSCache should be for internal
+   * user, so we can enforce specific expectations about its reference
+   * count: */
   assert (ps->for_internal_use);
+  assert (ps->ob_refcnt == 1);
+
   assert (ps->sql != NULL);
   assert (OBJECT_IS_VALID_TYPE_FOR_PREPARED_STATEMENT_SQL(ps->sql));
   assert (ps->cur != NULL);
@@ -269,14 +276,57 @@ static int _PreparedStatement_require_open(
    * Returns 0 if self was open; -1 if it was closed. */
   if (PreparedStatement_is_open(self)) { return 0; }
 
-  raise_exception(ProgrammingError,
-      failure_message != NULL ? failure_message :
-        "The PreparedStatement must be OPEN to perform this"
-        " operation."
-    );
+  if (self->state == PS_STATE_CONNECTION_TIMED_OUT) {
+    raise_exception(ConnectionTimedOut,
+        "This PreparedStatement's connection timed out, and"
+        " PreparedStatements cannot transparently survive a timeout."
+      );
+  } else {
+    raise_exception(ProgrammingError,
+        failure_message != NULL ? failure_message :
+          "The PreparedStatement must be OPEN to perform this"
+          " operation."
+      );
+  }
   return -1;
 } /* _PreparedStatement_require_open */
 
+static void PreparedStatement_create_references_to_superiors(
+    PreparedStatement *self, const boolean for_internal_use,
+    Cursor *cur
+  )
+{
+  assert (self != NULL);
+  assert (self->cur == NULL);
+  assert (cur != NULL);
+  /* The internal-use indicator that we've been provided should match the one
+   * set in the corresponding member of self: */
+  assert (for_internal_use == self->for_internal_use);
+
+  /* An non-internal (user-accessible) PreparedStatement owns a reference to
+   * its Cursor, while its Cursor does not own a reference to it.  Reference
+   * ownership for an internal PreparedStatement is just the opposite. */
+  if (!for_internal_use) {
+    Py_INCREF(cur);
+  }
+  self->cur = cur;
+} /* PreparedStatement_create_references_to_superiors */
+
+static void PreparedStatement_clear_references_to_superiors(
+    PreparedStatement *self
+  )
+{
+  assert (self != NULL);
+  {
+    Cursor *cur = self->cur;
+    assert (cur != NULL);
+    self->cur = NULL;
+    if (!self->for_internal_use) {
+      Py_DECREF(cur);
+    }
+  }
+} /* PreparedStatement_clear_references_to_superiors */
+
 static PreparedStatement *PreparedStatement_create(
     Cursor *cur, boolean for_internal_use
   )
@@ -292,13 +342,11 @@ static PreparedStatement *PreparedStatement_create(
   self->sql = NULL;
 
   assert (cur != NULL);
-  /* An non-internal (user-accessible) PreparedStatement owns a reference to
-   * its Cursor, while its Cursor does not own a reference to it.  Reference
-   * ownership for an internal PreparedStatement is just the opposite. */
-  if (!for_internal_use) {
-    Py_INCREF(cur);
-  }
-  self->cur = cur;
+  self->cur = NULL;
+  PreparedStatement_create_references_to_superiors(
+      self, for_internal_use, cur
+    );
+  assert (self->cur == cur);
 
   self->statement_type = NULL_STATEMENT_TYPE;
 
@@ -327,8 +375,19 @@ static int PreparedStatement_open(PreparedStatement* self,
     Cursor *cur, PyObject *sql
   )
 {
-  ISC_STATUS *sv = cur->status_vector;
-  CConnection *con = cur->connection;
+  ISC_STATUS *sv;
+  Transaction *trans;
+  CConnection *con;
+
+  assert (cur != NULL);
+  sv = cur->status_vector;
+
+  trans = cur->trans;
+  assert (trans != NULL);
+
+  con = Transaction_get_con(trans);
+  assert (con != NULL);
+
   CON_MUST_ALREADY_BE_ACTIVE(con);
 
   /* The caller should've already validated sql; sql should now be a non-NULL
@@ -361,8 +420,8 @@ static int PreparedStatement_open(PreparedStatement* self,
 
   /* Ask the database engine to compile the statement. */
   {
-    /* Note that we call CON_GET_TRANS_HANDLE_ADDR while holding the GIL. */
-    isc_tr_handle *trans_handle_addr = CON_GET_TRANS_HANDLE_ADDR(con);
+    /* Note that we call Transaction_get_handle_p while holding the GIL. */
+    isc_tr_handle *trans_handle_addr = Transaction_get_handle_p(trans);
     char *sql_raw_buffer = PyString_AS_STRING(sql);
     const Py_ssize_t sql_len = PyString_GET_SIZE(sql);
 
@@ -532,7 +591,7 @@ static int PreparedStatement_compare(
 
   if (    (a_sql == NULL || b_sql == NULL)
        || (a_cur == NULL || b_cur == NULL)
-       || (a_cur->connection != b_cur->connection)
+       || (a_cur->trans != b_cur->trans)
      )
   {
     /* Not equal. */
@@ -627,6 +686,7 @@ static int Cursor_clear_ps_description_tuples(Cursor *self) {
   /* Clears the description tuple of every PreparedStatement subordinate to
    * this Cursor. */
   PSCache *psc = &self->ps_cache_internal;
+
   if (!PSCache_has_been_deleted(psc)) {
     if (PSCache_traverse(psc, PSCacheMapped_clear_description_tuple) != 0)
     { goto fail; }
@@ -647,32 +707,28 @@ static int Cursor_clear_ps_description_tuples(Cursor *self) {
     return -1;
 } /* Cursor_clear_ps_description_tuples */
 
-static int CursorTrackerMapped_clear_description_tuples(
-    CursorTracker *node_prev, CursorTracker *node_cur
-  )
-{
-  /* Called on every element of a Connection's CursorTracker. */
-  Cursor *cur;
-  assert (node_cur != NULL);
-  cur = node_cur->contained;
-  assert (cur != NULL);
-  return Cursor_clear_ps_description_tuples(cur);
-} /* CursorTrackerMapped_clear_description_tuples */
-
 static int CConnection_clear_ps_description_tuples(
     CConnection *self
   )
 {
   /* Clears the description tuple of every PreparedStatement subordinate to
-   * this Connection (indirectly, by operating on the Connection's open
-   * Cursors). */
-  if (self->open_cursors != NULL) {
-    return CursorTracker_traverse(self->open_cursors,
-        CursorTrackerMapped_clear_description_tuples
-      );
-  } else {
-    return 0;
+   * this Connection. */
+  TransactionTracker *trans_node = self->transactions;
+  while (trans_node != NULL) {
+    Transaction *trans = trans_node->contained;
+    assert (trans != NULL);
+    {
+      CursorTracker *cur_node = trans->open_cursors;
+      while (cur_node != NULL) {
+        Cursor *cur = cur_node->contained;
+        assert (cur != NULL);
+        if (Cursor_clear_ps_description_tuples(cur) != 0) { return -1; }
+        cur_node = cur_node->next;
+      }
+    }
+    trans_node = trans_node->next;
   }
+  return 0;
 } /* CConnection_clear_ps_description_tuples */
 /* CLEAR-DESCRIPTION-TUPLE SUPPORT CODE : END */
 
@@ -690,8 +746,10 @@ static int PreparedStatement_isc_close(PreparedStatement *self,
    * in PreparedStatement_isc_drop) is sufficient. */
   ISC_STATUS *sv;
   assert (self->cur != NULL);
-  assert (self->cur->connection != NULL);
-  CON_MUST_ALREADY_BE_ACTIVE(self->cur->connection);
+  assert (self->cur->trans != NULL);
+  assert (Transaction_get_con(self->cur->trans) != NULL);
+
+  CON_MUST_ALREADY_BE_ACTIVE(Transaction_get_con(self->cur->trans));
 
   sv = self->cur->status_vector;
 
@@ -732,9 +790,10 @@ static int PreparedStatement_isc_drop(PreparedStatement *self,
 {
   ISC_STATUS *sv;
   assert (self->cur != NULL);
+  assert (self->cur->trans != NULL);
   #if (defined(ENABLE_CONNECTION_TIMEOUT) && !defined(NDEBUG))
   {
-    CConnection *con = self->cur->connection;
+    CConnection *con = Transaction_get_con(self->cur->trans);
     assert (con != NULL);
     if (Connection_timeout_enabled(con)) {
       assert (CURRENT_THREAD_OWNS_CON_TP(con));
@@ -754,13 +813,37 @@ static int PreparedStatement_isc_drop(PreparedStatement *self,
    * there's a syntax error.  In that case, self's state won't have been moved
    * to PS_STATE_OPEN. */
 
-  ENTER_GDAL
-  isc_dsql_free_statement(sv, &self->stmt_handle,
-      /* DSQL_drop means "free resources allocated for this statement; we're
-       * closing it permanently." */
-      DSQL_drop
-    );
-  LEAVE_GDAL
+  {
+    /* This code can be reached when the CTT is timing out a connection.  In
+     * that case, we want the GIL to remain held during the entire timeout
+     * operation. */
+    OPEN_LOCAL_GIL_MANIPULATION_INFRASTRUCTURE
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    const boolean should_manip_gil = NOT_RUNNING_IN_CONNECTION_TIMEOUT_THREAD;
+    if (should_manip_gil) {
+    #endif
+      LEAVE_GIL_WITHOUT_AFFECTING_DB_AND_WITHOUT_STARTING_CODE_BLOCK
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    }
+    #endif
+    ENTER_GDAL_WITHOUT_LEAVING_PYTHON
+
+    isc_dsql_free_statement(sv, &self->stmt_handle,
+        /* DSQL_drop means "free resources allocated for this statement; we're
+         * closing it permanently." */
+        DSQL_drop
+      );
+
+    LEAVE_GDAL_WITHOUT_ENTERING_PYTHON
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    if (should_manip_gil) {
+    #endif
+      ENTER_GIL_WITHOUT_AFFECTING_DB_AND_WITHOUT_ENDING_CODE_BLOCK
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    }
+    #endif
+    CLOSE_LOCAL_GIL_MANIPULATION_INFRASTRUCTURE
+  } /* end of lock manipulation scope */
 
   if (DB_API_ERROR(sv)) {
     raise_sql_exception(OperationalError,
@@ -825,7 +908,7 @@ static int PreparedStatement_close_without_unlink(PreparedStatement *self,
      * operations using self in the future. */
     if (self->cur->ps_current == self) {
       if (self->cur->state != CURSOR_STATE_CLOSED) {
-        if (Cursor_clear(self->cur, allowed_to_raise) != 0) { goto fail; }
+        Cursor_clear_and_leave_open(self->cur);
       }
       /* The lack of Py_DECREF here is deliberate, because the ps_current
        * member of Cursor never (conceptually) contains an owned reference. */
@@ -854,12 +937,31 @@ static int PreparedStatement_close_without_unlink(PreparedStatement *self,
     return -1;
 } /* PreparedStatement_close_without_unlink */
 
+static int PreparedStatement_untrack_with_superior_ref_clear_control(
+    PreparedStatement *self, const boolean allowed_to_raise,
+    const boolean clear_superior_refs
+  )
+{
+  if (PreparedStatement_close_without_unlink(self, allowed_to_raise) != 0) {
+    return -1;
+  }
+  assert (self->state == PS_STATE_DROPPED);
+  assert (self->cur != NULL);
+  if (clear_superior_refs) {
+    PreparedStatement_clear_references_to_superiors(self);
+    assert (self->cur == NULL);
+  }
+
+  return 0;
+} /* PreparedStatement_untrack_with_superior_ref_clear_control */
+
 static int PreparedStatement_untrack(PreparedStatement *self,
     boolean allowed_to_raise
   )
 {
-  /* Same as PreparedStatement_close_without_unlink. */
-  return PreparedStatement_close_without_unlink(self, allowed_to_raise);
+  return PreparedStatement_untrack_with_superior_ref_clear_control(
+      self, allowed_to_raise, TRUE
+    );
 } /* PreparedStatement_untrack */
 
 static int PreparedStatement_close_with_unlink(PreparedStatement *self,
@@ -892,14 +994,10 @@ static int PreparedStatement_close_with_unlink(PreparedStatement *self,
           SUPPRESS_EXCEPTION;
         }
       }
-
-      /* Note that the DECREF below applies only to PreparedStatements not for
-       * internal use (see notes in PreparedStatement_create and
-       * _Cursor_prepare_statement about reference ownership between
-       * PreparedStatements and Cursors). */
-      Py_DECREF(self->cur);
     }
-    self->cur = NULL;
+
+    PreparedStatement_clear_references_to_superiors(self);
+    assert (self->cur == NULL);
   }
 
   assert (allowed_to_raise ? self->state == PS_STATE_DROPPED : TRUE);
@@ -914,48 +1012,72 @@ static int PreparedStatement_close_with_unlink(PreparedStatement *self,
 
 /*********** PreparedStatement METHODS ACCESSIBLE TO PYTHON:BEGIN ************/
 static void pyob_PreparedStatement___del__(PreparedStatement *self) {
+  assert (
+        !self->for_internal_use
+      ? NOT_RUNNING_IN_CONNECTION_TIMEOUT_THREAD
+      : TRUE
+    );
+
   if (self->cur != NULL) {
-    CConnection *con = self->cur->connection;
+    Cursor *cur = self->cur;
+
+    /* We want to make sure that cur remains alive until we're done destroying
+     * self, but if this destructor is being called as a result of the
+     * execution of cur's destructor, we most definitely must not manipulate
+     * cur's reference count, which would cause cur to be "resurrected" and
+     * then for its destructor to execute again!
+     * Also, if self is for internal use, it never physically owns a reference
+     * to the cursor, so the reference count should not be manipulated. */
+    const boolean should_manipulate_cursor_refcnt = (
+        !self->for_internal_use && cur->ob_refcnt != 0
+      );
+
+    PyObject *con_python_wrapper = NULL;
+    CConnection *con;
+    assert (cur->trans != NULL);
+    con = Transaction_get_con(cur->trans);
+    assert (con != NULL);
+    con_python_wrapper = Transaction_get_con_python_wrapper(cur->trans);
+    assert (con_python_wrapper != NULL);
+
+    { /* hoop-jump for C90's retarded scoping */
+    const boolean needed_to_acquire_tp = !CURRENT_THREAD_OWNS_CON_TP(con);
+
+    /* Make sure con stays alive until we're done with it.  Note that cur
+     * might not stay alive until the end of this destructor. */
+    if (should_manipulate_cursor_refcnt) {
+      assert (cur->ob_refcnt != 0);
+      Py_INCREF(cur);
+    }
+    Py_INCREF(con);
+    Py_INCREF(con_python_wrapper);
 
     #ifdef ENABLE_CONNECTION_TIMEOUT
-    if (con == NULL || !Connection_timeout_enabled(con)) {
-    #endif
-      assert (con == NULL ? self->stmt_handle == NULL_STMT_HANDLE : TRUE);
-      if (PreparedStatement_close_with_unlink(self, TRUE) == 0) {
-        assert (self->cur == NULL);
-      } else {
-        SUPPRESS_EXCEPTION;
-      }
-    #ifdef ENABLE_CONNECTION_TIMEOUT
-    } else {
-      const boolean needed_to_acquire_tp = !CURRENT_THREAD_OWNS_CON_TP(con);
-
-      /* Make sure con stays alive until we're done with it.  Note that cur
-       * might not stay alive until the end of this destructor. */
-      Py_INCREF(con);
-
-      #ifdef ENABLE_CONNECTION_TIMEOUT
-      if (needed_to_acquire_tp) {
-        ACQUIRE_CON_TP_WITH_GIL_HELD(con);
-      }
-      #endif
-
-      if (PreparedStatement_close_with_unlink(self, TRUE) == 0) {
-        assert (self->cur == NULL);
-      } else {
-        SUPPRESS_EXCEPTION;
-      }
-
-      #ifdef ENABLE_CONNECTION_TIMEOUT
-      if (needed_to_acquire_tp) {
-        RELEASE_CON_TP(con);
-      }
-      #endif /* ENABLE_CONNECTION_TIMEOUT */
-
-      Py_DECREF(con);
+    if (needed_to_acquire_tp) {
+      ACQUIRE_CON_TP_WITH_GIL_HELD(con);
     }
     #endif
-  }
+
+    if (PreparedStatement_close_with_unlink(self, TRUE) == 0) {
+      assert (self->cur == NULL);
+    } else {
+      SUPPRESS_EXCEPTION;
+    }
+
+    #ifdef ENABLE_CONNECTION_TIMEOUT
+    if (needed_to_acquire_tp) {
+      RELEASE_CON_TP(con);
+    }
+    #endif /* ENABLE_CONNECTION_TIMEOUT */
+    } /* hoop-jump for C90's retarded scoping */
+
+    if (should_manipulate_cursor_refcnt) {
+      assert (cur->ob_refcnt != 0);
+      Py_DECREF(cur);
+    }
+    Py_DECREF(con);
+    Py_DECREF(con_python_wrapper);
+  } /* end of if (self->cur != NULL) block */
 
   /* Release the PreparedStatement struct itself: */
   PyObject_Del(self);
@@ -1010,20 +1132,22 @@ static PyObject *pyob_PreparedStatement_plan_get(
 {
   PyObject *ret = NULL;
 
+  PS_REQ_OPEN(self);
+
   assert (self->cur != NULL);
   #ifdef ENABLE_CONNECTION_TIMEOUT
     CUR_ACTIVATE__FORBID_TRANSPARENT_RESUMPTION(self->cur, return NULL);
   #endif /* ENABLE_CONNECTION_TIMEOUT */
 
-  PS_REQ_OPEN_WITH_FAILURE(self, goto fail);
-
-  assert (self->cur->connection != NULL);
-  CON_MUST_ALREADY_BE_ACTIVE(self->cur->connection);
+  assert (self->cur->trans != NULL);
+  assert (Transaction_get_con(self->cur->trans) != NULL);
+  CON_MUST_ALREADY_BE_ACTIVE(Transaction_get_con(self->cur->trans));
 
   ret = _generic_single_item_isc_dsql_sql_info_request(
       &self->stmt_handle, self->cur->status_vector, isc_info_sql_get_plan,
       1 /* Skip 1 byte (newline character at beginning of plan string. */
     );
+  if (ret == NULL) { goto fail; }
 
   goto clean;
   fail:
@@ -1036,7 +1160,7 @@ static PyObject *pyob_PreparedStatement_plan_get(
   clean:
     #ifdef ENABLE_CONNECTION_TIMEOUT
       CUR_PASSIVATE(self->cur);
-      CON_MUST_NOT_BE_ACTIVE(self->cur->connection);
+      CON_MUST_NOT_BE_ACTIVE(Transaction_get_con(self->cur->trans));
     #endif /* ENABLE_CONNECTION_TIMEOUT */
     return ret;
 } /* pyob_PreparedStatement_plan_get */
