@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of XSQLDA Utilities
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -140,8 +140,12 @@ static int reallocate_sqlda(XSQLDA **psqlda, boolean is_input_xsqlda,
 
 static int free_XSQLVAR_dynamically_allocated_memory(Cursor *cur) {
   /* Returns 0 if no error, -1 if error. */
-  PreparedStatement *ps = cur->ps_current;
-  XSQLDA *sqlda = ps->in_sqlda;
+  PreparedStatement *ps;
+  XSQLDA *sqlda;
+  assert (cur != NULL);
+  ps = cur->ps_current;
+  assert (ps != NULL);
+  sqlda = ps->in_sqlda;
 
   if (sqlda != NULL) {
     XSQLVAR *sqlvar;
