@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of Core
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -35,7 +35,6 @@
 
 /****************** "PRIVATE" DECLARATIONS:BEGIN *******************/
 static int Connection_require_open(CConnection *self, char *failure_message);
-static void Connection_clear_transaction_stats(CConnection *);
 static int CConnection_clear_ps_description_tuples(CConnection *);
 /****************** "PRIVATE" DECLARATIONS:END *******************/
 
@@ -98,23 +97,55 @@ static PyObject *py__make_output_translator_return_type_dict_from_trans_dict = N
 static PyObject *py_look_up_array_descriptor = NULL;
 static PyObject *py_look_up_array_subtype = NULL;
 static PyObject *pyob_Cursor_execute_exception_type_filter = NULL;
+static PyObject *pyob_validate_tpb = NULL;
+static PyObject *pyob_trans_info = NULL;
 
 /* These Python string objects are created when the module loads, then used by
  * multiple subsystems throughout the life of the process: */
 static PyObject *shared___s__C_con; /* _C_con */
+static PyObject *shared___s__main_trans; /* _main_trans */
+static PyObject *shared___s_ascii; /* ascii (canonical name of ASCII codec) */
+static PyObject *shared___s_charset; /* _charset */
+static PyObject *shared___s_DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP;
+static PyObject *shared___s_execute_immediate; /* execute_immediate */
+static PyObject *shared___s_strict; /* strict */
+
 /******************** GLOBAL VARIABLES:END ********************/
 
 /******************** IMPLEMENTATION:BEGIN ********************/
+PyObject *pyob_TrackerToList(AnyTracker *tracker);
+
+/* CConnection-related: */
 #ifdef ENABLE_CONNECTION_TIMEOUT
   static int Connection_close_from_CTT(volatile CConnection *con);
+
+  /* When a Connection times out, it cuts loose certain terminal objects
+   * (BlobReaders and PreparedStatements) which cannot under any circumstance
+   * be transparently resumed.  An attempt by the client programmer to use
+   * such an object should raise ConnectionTimedOut rather than
+   * ProgrammingError.  To implement that, we copy pointers to all BlobReaders
+   * and PreparedStatements attached to a Connection before we attempt to time
+   * the Connection out.  After timing it out, we then update the state
+   * indicator of the non-resumable objects so that they can raise
+   * ConnectionTimedOut if the client programmer attempts to use them again. */
   static BlobReader **Connection_copy_BlobReader_pointers(
       volatile CConnection *con, Py_ssize_t *count
     );
   static void Connection_former_BlobReaders_flag_timeout_and_free(
       BlobReader **blob_readers, Py_ssize_t count
     );
+
   #define Connection_former_BlobReader_pointers_free(blob_readers) \
     kimem_main_free(blob_readers)
+
+  static PreparedStatement **Connection_copy_PreparedStatement_pointers(
+      volatile CConnection *con, Py_ssize_t *count
+    );
+  static void Connection_former_PreparedStatements_flag_timeout_and_free(
+      PreparedStatement **prepared_statements, Py_ssize_t count
+    );
+  #define Connection_former_PreparedStatement_pointers_free(prepared_statements) \
+    kimem_main_free(prepared_statements)
 #endif /* ENABLE_CONNECTION_TIMEOUT */
 
 static int Connection_attach_from_members(CConnection *con
@@ -122,6 +153,44 @@ static int Connection_attach_from_members(CConnection *con
       , struct _ConnectionTimeoutParams *tp
     #endif
   );
+static PyObject *pyob_Connection_x_info(
+    boolean for_isc_database_info, isc_tr_handle *trans_handle_p,
+    PyObject *self, PyObject *args
+  );
+static boolean Connection_has_any_open_transaction(CConnection *con);
+
+/* Transaction-related: */
+boolean Transaction_is_main(Transaction *self) {
+  assert (self != NULL);
+  assert (self->con != NULL);
+  return (self->con->main_trans == self);
+} /* Transaction_is_main */
+static PyObject *pyob_Transaction_get_default_tpb(Transaction *self);
+
+typedef enum {
+  OP_COMMIT   =  1,
+  OP_ROLLBACK =  0
+} WhichTransactionOperation;
+
+typedef enum {
+  OP_RESULT_OK     =   0,
+  OP_RESULT_ERROR  =  -1
+} TransactionalOperationResult;
+
+static void Transaction_dist_trans_indicate_resultion(
+    Transaction *self, PyObject *group, const boolean is_resolved
+  );
+static int change_resolution_of_all_con_main_trans(
+    PyObject *group, PyObject *cons, const boolean is_resolved
+  );
+
+static TransactionalOperationResult Transaction_commit_or_rollback(
+    const WhichTransactionOperation op, Transaction *self,
+    const boolean retaining, const boolean allowed_to_raise
+  );
+
+/* BlobReader-related: */
+static int BlobReader_untrack(BlobReader *, boolean);
 
 /* These streaming blob support functions are declared here rather than in
  * _kinterbasdb.h so as to avoid compiler warnings that they're "declared but
@@ -131,11 +200,15 @@ static int validate_nonstandard_blob_config_dict(PyObject *,
   );
 static int BlobReaderTracker_release(BlobReaderTracker **);
 
+static boolean _check_statement_length(Py_ssize_t);
+
+static PyObject *pyob_Cursor_close(Cursor *self);
 static int _Cursor_require_open(Cursor *self, char *failure_message);
 static int Cursor_clear(Cursor *, boolean);
+static void Cursor_clear_and_leave_open(Cursor *);
 static int Cursor_ensure_PSCache(Cursor *self);
 static int Cursor_close_prepared_statements(Cursor *self,
-    boolean allowed_to_raise
+    const boolean allowed_to_raise, const boolean clear_ps_superior_refs
   );
 static int Cursor_close_without_unlink(Cursor *, boolean);
 static int Cursor_close_with_unlink(Cursor *, boolean);
@@ -143,7 +216,6 @@ static int Cursor_close_with_unlink(Cursor *, boolean);
 static int CursorTracker_release(CursorTracker **);
 
 typedef int (* CursorTrackerMappedFunction)(CursorTracker *, CursorTracker *);
-static int CursorTracker_traverse(CursorTracker *, CursorTrackerMappedFunction);
 static int Cursor_clear_ps_description_tuples(Cursor *);
 
 typedef int (* PSTrackerMappedFunction)(PSTracker *, PSTracker *);
@@ -157,10 +229,17 @@ static int PSTracker_traverse(PSTracker *, PSTrackerMappedFunction);
 #include "_kisupport_time.c"
 #include "_kinterbasdb_exception_functions.c"
 
+#include "_kicore_transaction_support.c"
+#include "_kicore_transaction_distributed.c"
 #include "_kicore_transaction.c"
 #include "_kiconversion.c"
 
 #ifdef ENABLE_CONNECTION_TIMEOUT
+  static TransactionalOperationResult
+    Connection_resolve_all_transactions_from_CTT(
+      volatile CConnection *con, const WhichTransactionOperation trans_op
+    );
+
   #include "_kicore_connection_timeout.c"
 #endif
 #include "_kicore_connection.c"
@@ -168,7 +247,7 @@ static int PSTracker_traverse(PSTracker *, PSTrackerMappedFunction);
 #include "_kicore_xsqlda.c"
 #include "_kicore_preparedstatement.c"
 #include "_kicore_cursor.c"
-#include "_kicore_execution.c"
+#include "_kicore_create_drop_db.c"
 
 #ifdef ENABLE_DB_EVENT_SUPPORT
   #include "_kinterbasdb_exception_functions_without_python.c"
@@ -179,12 +258,14 @@ static PyObject *pyob_provide_refs_to_python_entities(
     PyObject *self, PyObject *args
   )
 {
-  if (!PyArg_ParseTuple(args, "OOOOO",
+  if (!PyArg_ParseTuple(args, "OOOOOOO",
        &py_RowMapping_constructor,
        &py__make_output_translator_return_type_dict_from_trans_dict,
        &py_look_up_array_descriptor,
        &py_look_up_array_subtype,
-       &pyob_Cursor_execute_exception_type_filter
+       &pyob_Cursor_execute_exception_type_filter,
+       &pyob_validate_tpb,
+       &pyob_trans_info
      ))
   { return NULL; }
 
@@ -198,12 +279,18 @@ static PyObject *pyob_provide_refs_to_python_entities(
   REQ_CALLABLE(py__make_output_translator_return_type_dict_from_trans_dict);
   REQ_CALLABLE(py_look_up_array_descriptor);
   REQ_CALLABLE(py_look_up_array_subtype);
+  REQ_CALLABLE(pyob_Cursor_execute_exception_type_filter);
+  REQ_CALLABLE(pyob_validate_tpb);
+  REQ_CALLABLE(pyob_trans_info);
 
   /* We need these objects to live forever: */
   Py_INCREF(py_RowMapping_constructor);
   Py_INCREF(py__make_output_translator_return_type_dict_from_trans_dict);
   Py_INCREF(py_look_up_array_descriptor);
   Py_INCREF(py_look_up_array_subtype);
+  Py_INCREF(pyob_Cursor_execute_exception_type_filter);
+  Py_INCREF(pyob_validate_tpb);
+  Py_INCREF(pyob_trans_info);
 
   RETURN_PY_NONE;
 } /* pyob_provide_refs_to_python_entities */
@@ -311,6 +398,12 @@ static int init_shared_string_constants(void) {
     if (shared___s_ ## s == NULL) { goto fail; }
 
   INIT_SHARED_STRING_CONST(_C_con);
+  INIT_SHARED_STRING_CONST(_main_trans);
+  INIT_SHARED_STRING_CONST(ascii);
+  INIT_SHARED_STRING_CONST(charset);
+  INIT_SHARED_STRING_CONST(DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP);
+  INIT_SHARED_STRING_CONST(execute_immediate);
+  INIT_SHARED_STRING_CONST(strict);
 
   return 0;
 
@@ -334,30 +427,8 @@ static PyMethodDef kinterbasdb_GlobalMethods[] = {
   { "Connection_python_wrapper_obj_set",
     pyob_Connection_python_wrapper_obj_set,               METH_VARARGS      },
 
-  { "Connection_has_transaction",
-                            pyob_Connection_has_transaction,
-                                                          METH_VARARGS      },
-  { "Connection_clear_transaction_stats",
-                            pyob_Connection_clear_transaction_stats,
-                                                          METH_VARARGS      },
-
-  { "Connection_begin",     pyob_Connection_begin,        METH_VARARGS      },
-  { "Connection_prepare",   pyob_Connection_prepare,      METH_VARARGS      },
-  { "Connection_commit",    pyob_Connection_commit,       METH_VARARGS      },
-  { "Connection_rollback",  pyob_Connection_rollback,     METH_VARARGS      },
-
-  { "Connection_get_group", pyob_Connection_group_get,    METH_VARARGS      },
-  { "Connection_set_group", pyob_Connection_group_set,    METH_VARARGS      },
-  { "distributed_begin",    pyob_distributed_begin,       METH_VARARGS      },
-  { "distributed_prepare",  pyob_distributed_prepare,     METH_VARARGS      },
-  { "distributed_commit",   pyob_distributed_commit,      METH_VARARGS      },
-  { "distributed_rollback", pyob_distributed_rollback,    METH_VARARGS      },
-
   { "Connection_close",     pyob_Connection_close,        METH_VARARGS      },
 
-  { "Connection_execute_immediate",
-                            pyob_Connection_execute_immediate,
-                                                          METH_VARARGS      },
   { "Connection_database_info",
                             pyob_Connection_database_info,METH_VARARGS      },
   { "Connection_transaction_info",
@@ -387,6 +458,17 @@ static PyMethodDef kinterbasdb_GlobalMethods[] = {
                             pyob_Connection_dialect_set,  METH_VARARGS      },
   #endif /* INTERBASE_6_OR_LATER */
 
+  { "Connection_main_trans_get",
+                            pyob_Connection_main_trans_get,
+                                                          METH_VARARGS      },
+  { "Connection_transactions_get",
+                            pyob_Connection_transactions_get,
+                                                          METH_VARARGS      },
+
+  { "Connection_has_active_transaction",
+                            pyob_Connection_has_active_transaction,
+                                                          METH_VARARGS      },
+
   #ifdef ENABLE_DB_EVENT_SUPPORT
   { "EventConduit_create",  pyob_EventConduit_create,     METH_VARARGS      },
   #endif /* ENABLE_DB_EVENT_SUPPORT */
@@ -396,6 +478,14 @@ static PyMethodDef kinterbasdb_GlobalMethods[] = {
       METH_VARARGS
     },
 
+  /* StandaloneTransactionHandle methods: */
+  { "distributed_begin",    pyob_distributed_begin,       METH_VARARGS      },
+  { "distributed_prepare",  pyob_distributed_prepare,     METH_VARARGS      },
+  { "distributed_commit",   pyob_distributed_commit,      METH_VARARGS      },
+  { "distributed_rollback", pyob_distributed_rollback,    METH_VARARGS      },
+
+  /* Module-level functions that allow the Python layer to manipulate certain
+   * global variables within the C layer: */
   { "provide_refs_to_python_entities",
                             pyob_provide_refs_to_python_entities,
                                                           METH_VARARGS      },
@@ -424,13 +514,13 @@ static PyMethodDef kinterbasdb_GlobalMethods[] = {
   /* Connection_timeout_enabled should be available whether the feature is
    * compiled in or not: */
   { "Connection_timeout_enabled",
-                            pyob_Connection_timeout_enabled,
+                             pyob_Connection_timeout_enabled,
                                                           METH_VARARGS      },
 
   /* General utility functions: */
-  { "portable_int",         pyob_isc_portable_integer,    METH_VARARGS      },
+  { "portable_int",          pyob_isc_portable_integer,    METH_VARARGS     },
   { "raw_timestamp_to_tuple",
-                            pyob_raw_timestamp_to_tuple,  METH_VARARGS      },
+                             pyob_raw_timestamp_to_tuple,  METH_VARARGS     },
 
   /* The end: */
   { NULL,                NULL                                               }
@@ -477,23 +567,23 @@ static int init_kidb_exceptions(PyObject *d) {
 } /* init_kidb_exceptions */
 
 PyTypeObject ConnectionType = {
-  PyObject_HEAD_INIT(NULL)
+    PyObject_HEAD_INIT(NULL)
 
-  0,
-  "kinterbasdb.ConnectionType",
-  sizeof(CConnection),
-  0,
-  (destructor) pyob_Connection___del__,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0
-};
+    0,
+    "kinterbasdb.ConnectionType",
+    sizeof(CConnection),
+    0,
+    (destructor) pyob_Connection___del__,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0
+  };
 
 PyTypeObject BlobReaderType = { /* new-style class */
     PyObject_HEAD_INIT(NULL)
@@ -538,26 +628,26 @@ PyTypeObject BlobReaderType = { /* new-style class */
 
     0,                                  /* tp_alloc */
     0,                                  /* tp_new */
-};
+  };
 
-PyTypeObject TransactionHandleType = {
-  PyObject_HEAD_INIT(NULL)
+PyTypeObject StandaloneTransactionHandleType = {
+    PyObject_HEAD_INIT(NULL)
 
-  0,
-  "kinterbasdb.TransactionHandleType",
-  sizeof(TransactionHandleObject),
-  0,
-  (destructor) pyob_transaction_handle_del,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0,
-  0
-};
+    0,
+    "kinterbasdb.StandaloneTransactionHandle",
+    sizeof(StandaloneTransactionHandle),
+    0,
+    (destructor) StandaloneTransactionHandle___del__,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0,
+    0
+  };
 
 DL_EXPORT(void)
 init_kinterbasdb(void) {
@@ -570,7 +660,7 @@ init_kinterbasdb(void) {
 
   ConnectionType.ob_type = &PyType_Type;
   CursorType.ob_type = &PyType_Type;
-  TransactionHandleType.ob_type = &PyType_Type;
+  StandaloneTransactionHandleType.ob_type = &PyType_Type;
 
   #ifdef VERBOSE_DEBUGGING
   if (PyModule_AddIntConstant(m, "VERBOSE_DEBUGGING", 1) != 0) { goto fail; }
@@ -660,24 +750,6 @@ init_kinterbasdb(void) {
     return;
   }
 
-  /* DSR added null_connection when moving connection state detection from the
-   * Python level to the C level.  From the perspective of the Python-level
-   * kinterbasdb code, _kinterbasdb.null_connection is a null value like
-   * Python's None, except that it is of type ConnectionType instead of
-   * NoneType.
-   *   The kinterbasdb.Connection Python class can set its reference to its
-   * equivalent C type (self._C_conn) to _kinterbasdb.null_connection to
-   * indicate that the underlying connection is no longer valid.  Then the
-   * pyob_* functions in this C code that demand an argument of ConnectionType
-   * are satisfied by null_connection long enough to detect that it is not open
-   * (and that therefore the requested operation is not allowed). */
-  null_connection = Connection_create();
-  if (null_connection == NULL) {
-    PyErr_SetString(PyExc_ImportError, "Unable to create null_connection.");
-    return;
-  }
-  PyDict_SetItemString(d, "null_connection", (PyObject *) null_connection);
-
   if (init_kidb_exceptions(d) != 0) {
     PyErr_SetString(PyExc_ImportError,
         "Unable to initialize kinterbasdb exceptions."
@@ -719,9 +791,10 @@ init_kinterbasdb(void) {
     } \
   }
 
+  _INIT_C_TYPE_AND_SYS(Transaction, init_kidb_transaction);
+  _INIT_C_TYPE_AND_SYS(BlobReader, init_kidb_nonstandard_blob_support);
   _INIT_C_TYPE_AND_SYS(PreparedStatement, init_kidb_preparedstatement);
   _INIT_C_TYPE_AND_SYS(Cursor, init_kidb_cursor);
-  _INIT_C_TYPE_AND_SYS(BlobReader, init_kidb_nonstandard_blob_support);
   #ifdef ENABLE_DB_EVENT_SUPPORT
     _INIT_C_TYPE_AND_SYS(EventConduit, init_kidb_event_system);
 
@@ -739,6 +812,24 @@ init_kinterbasdb(void) {
     return;
   }
 
+  /* DSR added null_connection when moving connection state detection from the
+   * Python level to the C level.  From the perspective of the Python-level
+   * kinterbasdb code, _kinterbasdb.null_connection is a null value like
+   * Python's None, except that it is of type ConnectionType instead of
+   * NoneType.
+   *   The kinterbasdb.Connection Python class can set its reference to its
+   * equivalent C type (self._C_conn) to _kinterbasdb.null_connection to
+   * indicate that the underlying connection is no longer valid.  Then the
+   * pyob_* functions in this C code that demand an argument of ConnectionType
+   * are satisfied by null_connection long enough to detect that it is not open
+   * (and that therefore the requested operation is not allowed). */
+  null_connection = Connection_create();
+  if (null_connection == NULL) {
+    PyErr_SetString(PyExc_ImportError, "Unable to create null_connection.");
+    return;
+  }
+  PyDict_SetItemString(d, "null_connection", (PyObject *) null_connection);
+
   return;
 
   fail:
