@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Header File for Core
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -130,7 +130,7 @@ typedef unsigned char boolean;
 #endif
 
 #if NativeLongIsAtLeast64Bits
-  #define PythonIntOrLongFrom64BitValue(x) PyInt_FromLong((long) x)
+  #define PythonIntOrLongFrom64BitValue(x) PyInt_FromLong((long) (x))
 #else
   #define PythonIntOrLongFrom64BitValue(x) \
     ( ((x) < LONG_MIN || (x) > LONG_MAX) ? PyLong_FromLongLong(x) : PyInt_FromLong((long) (x)) )
@@ -213,6 +213,20 @@ const Py_ssize_t PY_SSIZE_T_MAX__CIRCUMVENT_COMPILER_COMPLAINT = PY_SSIZE_T_MAX;
 
 #if (defined(METADATALENGTH) && defined(SQLDA_VERSION2) && defined(blr_boolean_dtype))
   #define INTERBASE_7_OR_LATER
+#else
+  /* Provide some constants so that versions of KInterbasDB compiled against
+   * Firebird can still work at least minimally with Interbase 7+.  The FB
+   * developers will probably maintain compatibility with these constants when
+   * FB adds more types, so compatibility shouldn't be a problem. */
+  #ifndef ISC_BOOLEAN
+    #define ISC_BOOLEAN signed short
+  #endif
+  #ifndef SQL_BOOLEAN
+    #define SQL_BOOLEAN 590
+  #endif
+  #ifndef blr_boolean_dtype
+    #define blr_boolean_dtype 17
+  #endif
 #endif
 
 /* Beginning with Firebird 1.5, Firebird defines an integer constant FB_API_VER
@@ -383,7 +397,6 @@ typedef enum {
 #define DB_API_ERROR(status_vector) \
   ( (((status_vector)[0] == 1) && (status_vector)[1] > 0) ? TRUE : FALSE )
 
-
 #define MIN(a, b) ( (a < b) ? a : b )
 #define MAX(a, b) ( (a > b) ? a : b )
 
@@ -391,6 +404,14 @@ typedef enum {
   Py_INCREF(Py_None); \
   return Py_None;
 
+#define HANDLE_ERROR_WHEN_POSSIBLY_NOT_ALLOWED_TO_RAISE(allowed_to_raise) \
+  if (allowed_to_raise) { \
+    goto fail; \
+  } else { \
+    status = -1; \
+    SUPPRESS_EXCEPTION; \
+  }
+
 /******************** HANDY ABBREVIATIONS:END ********************/
 
 /******************** MODULE TYPE DEFINITIONS:BEGIN ********************/
@@ -398,13 +419,113 @@ typedef enum {
 /* See Appendix B of Interbase 6 API Guide for explanation of Interbase API
  * structures. */
 
+typedef struct _AnyTracker {
+  void *contained;
+  struct _AnyTracker *next;
+} AnyTracker;
+
+extern PyTypeObject TransactionType;
 extern PyTypeObject BlobReaderType;
 extern PyTypeObject ConnectionType;
 extern PyTypeObject CursorType;
-extern PyTypeObject TransactionHandleType;
+extern PyTypeObject StandaloneTransactionHandleType;
 extern PyTypeObject PreparedStatementType;
 extern PyTypeObject EventConduitType;
 
+struct _BlobReaderTracker;
+struct _CursorTracker;
+struct _CConnection;
+
+typedef enum {
+  TR_STATE_CREATED,
+  TR_STATE_UNRESOLVED,
+  TR_STATE_RESOLVED,
+  TR_STATE_CLOSED,
+  TR_STATE_CONNECTION_TIMED_OUT
+} TransactionState;
+
+typedef struct { /* Transaction */
+  PyObject_HEAD /* Python API - infrastructural macro. */
+
+  TransactionState state;
+
+  /* Will be NULL if this Transaction has been close()d: */
+  struct _CConnection *con;
+
+  /* We maintain references to the CConnection instance and its Python wrapper
+   * object (an instance of Python class kinterbasdb.Connection) separately.
+   * The CConnection is what we actually use, but we need to ensure that the
+   * Python wrapper object sticks around so that the CConnection remains
+   * functional. */
+  PyObject *con_python_wrapper;
+
+  isc_tr_handle trans_handle; /* database API - transaction handle. */
+  PyObject *group; /* distributed transaction group */
+
+  PyObject *default_tpb;
+
+  /* open_cursors serves the same purpose as open_blobreaders, but for Cursors.
+   * See the documentation of open_blobreaders (below) for an overview. */
+  struct _CursorTracker *open_cursors;
+
+  /* Documentation for member open_blobreaders:
+   *
+   * The database server requires that open blob handles be closed explicitly,
+   * lest there be leakage.  Many users of CPython, however, are used to
+   * relying on the deterministic, reference-count-based finalization to close
+   * open resources for them implicitly, as in:
+   *   file('/blah/blah', 'rb').read()
+   *
+   * All operations on blob handles require an active database handle and
+   * transaction handle, so it is clearly necessary to close any open
+   * BlobReaders before resolving a transaction, lest those BlobReaders become
+   * uncloseable and cause memory leakage in the server process.
+   *
+   * BlobReader_open adds the new BlobReader to the Transaction's tracker, and
+   * BlobReader_close_with_unlink (which is called by the BlobReader destructor
+   * and also explicitly callable by the client programmer) removes the
+   * BlobReader from that tracker.
+   *
+   * However, if the transaction is about to be resolved, any BlobReaders that
+   * have not yet been garbage collected or explicitly closed need to be closed
+   * before the transaction ends.
+   *
+   * The obvious data structure with which a Transaction could track its open
+   * BlobReaders is the Python dictionary, but that would create a cycle.
+   * A BlobReader already maintains a reference to its Transaction, so it is
+   * guaranteed that a BlobReader's Transaction will outlive the BlobReader.
+   * If each BlobReader were added to a Python dictionary owned by the
+   * Transaction, there'd be be a cycle.
+   *
+   * Python's garbage collector has special facilities for handling cyclic
+   * garbage, but I'd rather not use them because they violate the determinism
+   * that most CPython users expect from their garbage collector.  So I opted
+   * for a C linked list as the data structure of the Transaction's tracker.
+   * The number of BlobReaders tracked at any given time is likely to be very
+   * small, so the poor lookup performance of the linked list is not likely to
+   * cause significant performance degradation. */
+  struct _BlobReaderTracker *open_blobreaders;
+
+  LONG_LONG n_physical_transactions_started;
+  LONG_LONG n_prepared_statements_executed_since_current_phys_start;
+} Transaction;
+
+typedef struct _TransactionTracker {
+  Transaction *contained;
+  struct _TransactionTracker *next;
+} TransactionTracker;
+
+/* Note the distinction between a Transaction object being:
+ *  - "not closed" (that is, its close() method hasn't been called)
+ * and
+ *  - "active" (that is, a physical transaction has been started and not yet
+ *    resolved) */
+#define Transaction_is_not_closed(self) ((self)->state < TR_STATE_CLOSED)
+#define Transaction_is_active(self) ((self)->state == TR_STATE_UNRESOLVED)
+
+#define Transaction_con_timed_out(self) \
+  ((self)->state == TR_STATE_CONNECTION_TIMED_OUT)
+
 /* This structure supports a moderately ugly approach to determining the
  * precision of a given field (by querying the system tables--see
  * _kiconversion_field_precision.c). */
@@ -430,14 +551,14 @@ typedef struct {
 #ifdef ENABLE_CONNECTION_TIMEOUT
   struct _ConnectionTimeoutParams;
 #endif
-struct _BlobReaderTracker;
-struct _CursorTracker;
 
 typedef enum {
   CON_STATE_CLOSED = 0,
   CON_STATE_OPEN = 1
 } ConnectionState;
 
+#define Connection_is_closed(con) ((con)->state == CON_STATE_CLOSED)
+
 typedef struct _CConnection { /* definition of type CConnection */
   PyObject_HEAD /* Python API - infrastructural macro. */
 
@@ -458,8 +579,9 @@ typedef struct _CConnection { /* definition of type CConnection */
   unsigned short   dialect;
   isc_db_handle    db_handle; /* Interbase API - database handle. */
 
-  isc_tr_handle    trans_handle; /* Interbase API - transaction handle. */
-  PyObject *group; /* Distributed transaction group. */
+  /* Added separate Transaction class in January 2007 for KIDB 3.3: */
+  Transaction *main_trans;
+  struct _TransactionTracker *transactions;
 
   /* Buffer used by Interbase API to store error status of calls. */
   ISC_STATUS       status_vector[STATUS_VECTOR_SIZE];
@@ -468,10 +590,6 @@ typedef struct _CConnection { /* definition of type CConnection */
   CursorDescriptionCache *desc_cache;
   #endif
 
-  struct _CursorTracker *open_cursors;
-
-  LONG_LONG n_prepared_statements_executed_since_trans_start;
-
   /* 2005.06.21: */
   /* Nested Python dictionary of the following form:
    *   {'RELATION_NAME_0': {'FIELD_NAME_0': characterSetId, ...}, ...}
@@ -485,45 +603,6 @@ typedef struct _CConnection { /* definition of type CConnection */
 
   PyObject *output_type_trans_return_type_dict;
 
-  /* Documentation for member open_blobreaders:
-   *
-   * The database server requires that open blob handles be closed explicitly,
-   * lest there be leakage.  Many users of CPython, however, are used to
-   * relying on the deterministic, reference-count-based finalization to close
-   * open resources for them implicitly, as in:
-   *   file('/blah/blah', 'rb').read()
-   *
-   * All operations on blob handles require an active database handle and
-   * transaction handle, both of which are managed by the Connection in
-   * kinterbasdb.  So it is clearly necessary to close any open BlobReaders
-   * before resolving a Connection's transaction, lest those BlobReaders
-   * become uncloseable and cause memory leakage in the server process.
-   *
-   * BlobReader_open adds the new BlobReader to the Connection's tracker, and
-   * BlobReader_close_with_unlink (which is called by the BlobReader destructor
-   * and also explicitly callable by the client programmer) removes the
-   * BlobReader from that tracker.
-   *
-   * However, if the Connection's transaction is about to be resolved, any
-   * BlobReaders that have not yet been garbage collected or explicitly closed
-   * need to be closed before the transaction ends.
-   *
-   * The obvious data structure with which a Connection could track its open
-   * BlobReaders is the Python dictionary, but that would create a cycle.
-   * A BlobReader already maintains a reference to its Connection, so it is
-   * guaranteed that a BlobReader's Connection will outlive the BlobReader.  If
-   * each BlobReader were added to a Python dictionary owned by the Connection,
-   * there'd be be a cycle.
-   *
-   * Python's garbage collector has special facilities for handling cyclic
-   * garbage, but I'd rather not use them because they violate the determinism
-   * that most CPython users expect from their garbage collector.  So I opted
-   * for a C linked list as the data structure of the Connection's tracker.
-   * The number of BlobReaders tracked at any given time is likely to be very
-   * small, so the poor lookup performance of the linked list is not likely to
-   * cause significant performance degradation. */
-  struct _BlobReaderTracker *open_blobreaders;
-
   #ifdef ENABLE_CONNECTION_TIMEOUT
     struct _ConnectionTimeoutParams *timeout;
     #define Connection_timeout_enabled(con) \
@@ -565,10 +644,8 @@ typedef struct { /* definition of type BlobReader */
 
   BlobReaderState state;
 
-  CConnection *con;
-  /* We maintain references to the CConnection instance and its Python
-   * wrapper object (an instance of kinterbasdb.Connection) separately.  This
-   * prevents the CConnection from being prematurely garbage collected. */
+  Transaction *trans;
+  /* TAG:TRANSACTION_SUBORDINATE_OBJECT_SURVIVAL_BYPASS: */
   PyObject *con_python_wrapper;
 
   isc_blob_handle blob_handle;
@@ -598,30 +675,15 @@ typedef struct _BlobReaderTracker {
 #define BlobReader_con_timed_out(self) \
   ((self)->state == BLOBREADER_STATE_CONNECTION_TIMED_OUT)
 
-#define CLEAR_OPEN_BLOBREADER_LIST_IF_NECESSARY(con, on_failure) \
-  if ((con)->open_blobreaders != NULL) { \
-    if (BlobReaderTracker_release(&(con)->open_blobreaders) != 0) { \
-      on_failure; \
-    } \
-    assert ((con)->open_blobreaders == NULL); \
-  }
-
-#define CLEAR_OPEN_CURSOR_LIST_IF_NECESSARY(con, on_failure) \
-  if ((con)->open_cursors != NULL) { \
-    if (CursorTracker_release(&(con)->open_cursors) != 0) { \
-      on_failure; \
-    } \
-    assert ((con)->open_cursors == NULL); \
-  }
-
 #define OBJECT_IS_VALID_TYPE_FOR_PREPARED_STATEMENT_SQL(o) \
   ((o) != NULL && PyString_CheckExact((o)))
 
 typedef enum {
   PS_STATE_CREATED = 0, /* Before isc_dsql_allocate_statement, etc. */
-  PS_STATE_OPEN = 1,
-  PS_STATE_CLOSED = 2,  /* After isc_dsql_free_statement(..., DSQL_close) */
-  PS_STATE_DROPPED = 3  /* After isc_dsql_free_statement(..., DSQL_drop) */
+  PS_STATE_OPEN,
+  PS_STATE_CLOSED,  /* After isc_dsql_free_statement(..., DSQL_close) */
+  PS_STATE_DROPPED,  /* After isc_dsql_free_statement(..., DSQL_drop) */
+  PS_STATE_CONNECTION_TIMED_OUT
 } PreparedStatementState;
 
 typedef struct {
@@ -719,18 +781,16 @@ typedef enum {
   CURSOR_STATE_DROPPED = 3
 } CursorState;
 
+#define Cursor_is_open(cur) ((cur)->state <= CURSOR_STATE_OPEN)
+
 typedef struct _Cursor { /* definition of type Cursor */
   PyObject_HEAD /* Python API - infrastructural macro. */
 
   CursorState state;
 
-  /* Connection associated with this cursor (an instance of a C type): */
-  CConnection *connection;
-  /* Python wrapper around the C CConnection instance associated with this
-   * cursor.  A reference to the Python wrapper object must be retained so that
-   * the Python wrapper object will not cause the CConnection to be
-   * deleted while this cursor still might need to use it. */
-  PyObject *connection_python_wrapper;
+  Transaction *trans;
+  /* TAG:TRANSACTION_SUBORDINATE_OBJECT_SURVIVAL_BYPASS: */
+  PyObject *con_python_wrapper;
 
   PreparedStatement *ps_current;
   /* See notes at declarations of these types.  A given PreparedStatement will
@@ -765,7 +825,8 @@ typedef struct _Cursor { /* definition of type Cursor */
    * it in the Pythonic way, instead of calling isc_dsql_fetch and encountering
    * an error. */
   ISC_STATUS last_fetch_status;
-  #define NO_FETCH_ATTEMPTED_YET -1
+  #define NO_FETCH_ATTEMPTED_YET   -1
+  #define RESULT_SET_EXHAUSTED    100
 
   /* Dynamic type translation support fields: */
   PyObject *type_trans_in;
@@ -794,11 +855,14 @@ typedef struct teb {
   char *tpb_ptr;
 } ISC_TEB;
 
-typedef struct { /* definition of type TransactionHandleObject */
+typedef struct { /* definition of type StandaloneTransactionHandle */
   PyObject_HEAD /* Python API - infrastructural macro. */
 
   isc_tr_handle native_handle;
-} TransactionHandleObject;
+} StandaloneTransactionHandle;
+
+#define StandaloneTransactionHandle_Check(obj) \
+  ((obj)->ob_type == &StandaloneTransactionHandleType)
 
 
 /******************** MODULE TYPE DEFINITIONS:END ********************/
