@@ -1,6 +1,6 @@
 /* KInterbasDB Python Package - Implementation of "Exception-Raising" Functions
  *
- * Version 3.2
+ * Version 3.3
  *
  * The following contributors hold Copyright (C) over their respective
  * portions of code (see license.txt for details):
@@ -9,7 +9,7 @@
  *   1998-2001 [alex]  Alexander Kuznetsov   <alexan@users.sourceforge.net>
  * [Maintainers (after version 2.0-0.3.1):]
  *   2001-2002 [maz]   Marek Isalski         <kinterbasdb@maz.nu>
- *   2002-2006 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
+ *   2002-2007 [dsr]   David Rushby          <woodsplitter@rocketmail.com>
  * [Contributors:]
  *   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
  *   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
@@ -204,7 +204,8 @@ void raise_sql_exception_exc_type_filter(
         }
 
         /* Create filter arguments: */
-        py_raw_code = PyInt_FromLong(raw_code);
+        /* 2007.02.10: FB 2.1 fix: ISC_STATUS can be > INT_MAX in FB 2.1: */
+        py_raw_code = PythonIntOrLongFrom64BitValue(raw_code);
         if (py_raw_code == NULL) { goto exception_type_filter_finish; }
 
         py_sql_code = PyInt_FromLong(db_error_code);
@@ -331,14 +332,16 @@ void raise_sql_exception(
 }
 
 static void raise_exception_with_numeric_error_code(
-    PyObject *exc_type, long error_code, const char *description
+    PyObject *exc_type,
+    const LONG_LONG error_code, /* 2007.02.10: FB 2.1 fix: long->LONG_LONG */
+    const char *description
   )
 {
   /* raise_exception_with_numeric_error_code allows the database API error code
    * to be set directly, rather than having it extracted from a status vector.
    * Thus, raise_exception_with_numeric_error_code might be said to "fall
    * midway between raise_sql_exception and raise_exception". */
-  PyObject *exceptionTuple = Py_BuildValue("(ls)", error_code, description);
+  PyObject *exceptionTuple = Py_BuildValue("(Ls)", error_code, description);
   if (exceptionTuple == NULL) { return; }
   PyErr_SetObject(exc_type, exceptionTuple);
   Py_DECREF(exceptionTuple);
