@@ -14,14 +14,18 @@
 #   2001      [eac]   Evgeny A. Cherkashin  <eugeneai@icc.ru>
 #   2001-2002 [janez] Janez Jere            <janez.jere@void.si>
 
+# This module can be conveniently activated as the process-wide default via:
+#   kinterbasdb.init(type_conv=100)
+
 from kinterbasdb import typeconv_datetime_stdlib
+# fixedpoint is a third-party module, not part of the standard library:
 from kinterbasdb import typeconv_fixed_fixedpoint
 from kinterbasdb import typeconv_text_unicode
 
 _underlying_modules = (
     typeconv_datetime_stdlib,
     typeconv_fixed_fixedpoint,
-    typeconv_text_unicode # 2004.01.03
+    typeconv_text_unicode
   )
 
 # Load the required members from the underlying modules into the namespace of
