@@ -1,6 +1,6 @@
 # KInterbasDB Python Package - Type Conv : Text/Unicode
 #
-# Version 3.2
+# Version 3.3
 #
 # The following contributors hold Copyright (C) over their respective
 # portions of code (see license.txt for details):
@@ -89,6 +89,68 @@ DB_TO_PYTHON_ENCODING_MAP = {
            65: 'cp1258',           #: 'WIN1258' (Firebird 2.0+)
   }
 
+DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP = {
+    # DB CHAR SET NAME    :   PYTHON CODEC NAME (CANONICAL)
+    # ---------------------------------------------------------------------------
+    'OCTETS'              :   None, # Allow to pass through unchanged.
+    'UNICODE_FSS'         :   'utf_8',
+    'UTF8'                :   'utf_8', # (Firebird 2.0+)
+    'SJIS_0208'           :   'shift_jis',
+    'EUCJ_0208'           :   'euc_jp',
+    'DOS737'              :   'cp737',
+    'DOS437'              :   'cp437',
+    'DOS850'              :   'cp850',
+    'DOS865'              :   'cp865',
+    'DOS860'              :   'cp860',
+    'DOS863'              :   'cp863',
+    'DOS775'              :   'cp775',
+    'DOS862'              :   'cp862',
+    'DOS864'              :   'cp864',
+    'ISO8859_1'           :   'iso8859_1',
+    'ISO8859_2'           :   'iso8859_2',
+    'ISO8859_3'           :   'iso8859_3',
+    'ISO8859_4'           :   'iso8859_4',
+    'ISO8859_5'           :   'iso8859_5',
+    'ISO8859_6'           :   'iso8859_6',
+    'ISO8859_7'           :   'iso8859_7',
+    'ISO8859_8'           :   'iso8859_8',
+    'ISO8859_9'           :   'iso8859_9',
+    'ISO8859_13'          :   'iso8859_13',
+    'KSC_5601'            :   'euc_kr',
+    'DOS852'              :   'cp852',
+    'DOS857'              :   'cp857',
+    'DOS861'              :   'cp861',
+    'DOS866'              :   'cp866',
+    'DOS869'              :   'cp869',
+    'WIN1250'             :   'cp1250',
+    'WIN1251'             :   'cp1251',
+    'WIN1252'             :   'cp1252',
+    'WIN1253'             :   'cp1253',
+    'WIN1254'             :   'cp1254',
+    'BIG_5'               :   'big5',
+    'GB_2312'             :   'gb2312',
+    'WIN1255'             :   'cp1255',
+    'WIN1256'             :   'cp1256',
+    'WIN1257'             :   'cp1257',
+    'KOI8-R'              :   'koi8_r', # (Firebird 2.0+)
+    'KOI8-U'              :   'koi8_u', # (Firebird 2.0+)
+    'WIN1258'             :   'cp1258', # (Firebird 2.0+)
+  }
+
+# The inverse of DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP:
+PYTHON_ENCODING_TO_DB_CHAR_SET_NAME_MAP = dict(
+        (val, key)
+    for (key, val)
+    in DB_CHAR_SET_NAME_TO_PYTHON_ENCODING_MAP.iteritems()
+  )
+
+DB_CHAR_SET_NAME_TO_DB_CHAR_SET_ID_MAP = {} # Example entry: 'WIN1251': 52
+for dbCharSetID, pythonEncodingName in DB_TO_PYTHON_ENCODING_MAP.iteritems():
+    dbCharSetName = PYTHON_ENCODING_TO_DB_CHAR_SET_NAME_MAP[pythonEncodingName]
+    DB_CHAR_SET_NAME_TO_DB_CHAR_SET_ID_MAP[dbCharSetName] = dbCharSetID
+del dbCharSetID, pythonEncodingName, dbCharSetName
+
+
 _UNKNOWN_CHARSET_MSG = (
     "Don't know how to %s value %s charset with numeric ID %d."
     " If you are using an unofficial character set, you should add a"
